# Shallot is a Qt based file manager.
# Copyright (C) 2014-2021 Josef Hahn
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyPep8Naming
import xml.etree.cElementTree as ET

# noinspection PyUnresolvedReferences
import shallot
import json
import re
import subprocess
import threading
import time
import traceback


Strings = shallot.IntlStringMap(
    Cancel={"en": "Cancel", "de": "Abbrechen"},
    AniseProjectName={"en": "Anise project name", "de": "Anise Projektname"},
    ProjectNameNone={"en": "(none)", "de": "(keiner)"},
    RunAnAniseTask={"en": "Run an Anise task", "de": "Anise-Task ausführen"},
    TaskInParentPattern={"en": "{taskname} (in {parentname})", "de": "{taskname} (in {parentname})"},
    TasksMore={"en": "more", "de": "mehr"},
)


class Utils:

    lock = threading.Lock()

    isanisesyncedcache = {}

    aniseprojectdetailscache = {}

    aniseprojecttaskscache = {}

    @staticmethod
    def clearcaches():
        with Utils.lock:
            Utils.isanisesyncedcache.clear()
            Utils.aniseprojectdetailscache.clear()
            Utils.aniseprojecttaskscache.clear()

    @staticmethod
    def _findaniseroot(op, eurl) -> object:
        seurl = eurl.as_string()
        fromcache = Utils.isanisesyncedcache.get(seurl, None)
        rooteurl = None
        projectdescfileeurl = None
        if fromcache:
            lifetime, rooteurl, projectdescfileeurl = fromcache
            if lifetime < time.time():
                fromcache = None
        if fromcache:
            return op, shallot.Eurl.from_string(rooteurl) if rooteurl else False, \
                   shallot.Eurl.from_string(projectdescfileeurl) if projectdescfileeurl else False  # might be 'False'
        if len(Utils.isanisesyncedcache) > 50:
            sortlst = sorted(Utils.isanisesyncedcache.items(), key=lambda x: x[1][0])
            for elm in sortlst[:10]:
                Utils.isanisesyncedcache.pop(elm[0])
        if eurl.path() != "/":
            _op, parentroot, parentpdesc = Utils._findaniseroot(op, eurl.parent_segment())
            if parentroot:
                Utils.isanisesyncedcache[seurl] = (time.time() + 200, parentroot.as_string(), parentpdesc.as_string())
                return op, parentroot, parentpdesc
        if op is None:
            op = shallot.Operations.create()
        rtpdesc = None
        for srtpdescalt in ["_meta/_projectdesc", "_projectdesc"]:
            esrtpdescalt = eurl.with_appended_segments(srtpdescalt)
            if op.filesystem().get_type(esrtpdescalt) == shallot.Filesystem.NodeType.File:
                rtpdesc = esrtpdescalt
                break
        if rtpdesc:
            Utils.isanisesyncedcache[seurl] = (time.time() + 600, seurl, rtpdesc.as_string())
            return op, eurl, rtpdesc
        else:
            Utils.isanisesyncedcache[seurl] = (time.time() + 200, False, False)
            return op, False, False

    @staticmethod
    def isanisesynced(op, eurl) -> object:
        if eurl.scheme() == "file":
            with Utils.lock:
                return Utils._findaniseroot(op, eurl)
        else:
            return op, False, False

    @staticmethod
    def getaniseprojectname(op, projectdesceurl) -> object:
        sprojectdesceurl = projectdesceurl.as_string()
        with Utils.lock:
            projdets = None
            fromcache = Utils.aniseprojectdetailscache.get(sprojectdesceurl, None)
            if fromcache:
                lifetime, projdets = fromcache
                if lifetime < time.time():
                    projdets = None
            if projdets is None:
                projdets = {}
                Utils.aniseprojectdetailscache[sprojectdesceurl] = (time.time() + 200, projdets)
            projname = projdets.get("name", None)
            if projname is None:
                pdcont = projdets.get("__content", None)
                if pdcont is None:
                    try:
                        pdcont = op.filesystem().get_file_content(projectdesceurl).readall()
                    except Exception:
                        shallot.Logging.log_error(traceback.format_exc())
                        pdcont = ""
                    projdets["__content"] = pdcont
                xpddoc = ET.ElementTree(ET.fromstring(pdcont))
                xpdroot = xpddoc.getroot()
                for xentry in xpdroot.getchildren():
                    if xentry.tag == "val":
                        if xentry.attrib["k"] == "name":
                            projname = xentry.attrib["v"]
                            break
                projdets["name"] = projname
            return projname

    @staticmethod
    def getaniseprojecttasks(rooteurl) -> object:
        srooteurl = rooteurl.as_string()
        with Utils.lock:
            projtasks = None
            fromcache = Utils.aniseprojecttaskscache.get(srooteurl, None)
            if fromcache:
                lifetime, projtasks = fromcache
                if lifetime < time.time():
                    projtasks = None
            if projtasks is None:
                try:
                    tlrta = subprocess.check_output(["anise", "base.getregisteredtasks", "--no-gui"],
                                                    cwd=rooteurl.path()).decode()
                    m = re.search("TASK RESULT:\s*(\S+)\s*", tlrta)
                    if m:
                        ftlrt = m.group(1)
                        with open(ftlrt, "rb") as f:
                            btlrt = f.read()
                except Exception:
                    btlrt = b"[]"
                    shallot.Logging.log_error(traceback.format_exc())
                projtasks = json.loads(btlrt)
                Utils.aniseprojecttaskscache[srooteurl] = (time.time() + 200, projtasks)
            return projtasks


def exec_action(cls, *a, **b):
    a = cls(*a, **b)
    a.initialize_sync()
    a.execute()


class AnisePanelDetailProjectName(shallot.PanelDetails.SingleSelectionPanelDetail):
    def __init__(self):
        shallot.PanelDetails.SingleSelectionPanelDetail.__init__(
            self, positionGroup=shallot.PanelDetails.PanelDetail.INDEX_EXOTIC, positionIndex=27599, valueWidthHint=5)

    def set_value(self, detail, node, op):
        op, res, pdfile = Utils.isanisesynced(op, node.eurl())
        if res:
            projectname = []
            _pn = Utils.getaniseprojectname(op, pdfile)
            _pn = _pn or Strings.ProjectNameNone
            projectname.append(shallot.PanelDetails.PanelDetailValueElementString(_pn))
            return [(Strings.AniseProjectName, projectname)]


class RunAniseTasksActionMenu(shallot.Actions.SubmenuAction):

    def __init__(self, nodes):
        super().__init__(Strings.RunAnAniseTask, icon="media-playback-start")
        self.eurl = nodes[0].eurl()

    def initialize(self):
        op, res, pdfile = Utils.isanisesynced(None, self.eurl)
        if res:
            tasks = [t for t in Utils.getaniseprojecttasks(res) if not t['invisible']]
            importanttasks = [t for t in tasks if t['important']]
            unimportanttasks = [t for t in tasks if not t['important']]
            repath = res.path()
            subitems = list([RunAniseTaskAction(repath, ts) for ts in importanttasks])
            if unimportanttasks:
                unimportantsubitems = list([RunAniseTaskAction(repath, ts) for ts in unimportanttasks])
                subitems.append(RunAniseTasksMoreActionMenu(unimportantsubitems))
            self.set_subitems(subitems)
        else:
            self.set_visible(False)


class RunAniseTasksMoreActionMenu(shallot.Actions.SubmenuAction):

    def __init__(self, subitems):
        super().__init__(Strings.TasksMore, icon="arrow-down")
        self.set_subitems(subitems)


class RunAniseTaskAction(shallot.Actions.ActionAction):

    def __init__(self, rootpath, task):
        self.rootpath = rootpath
        if task['label']:
            label = task['label']
        else:
            if task['parentname']:
                label = Strings.TaskInParentPattern.format(**task)
            else:
                label = task['taskname']
        self.taskname = task['name']
        super().__init__(label)

    def action(self, info):
        info.set_manual_intervention_needed(False)
        try:
            subprocess.check_output(["anise", self.taskname], cwd=self.rootpath)
        except Exception:
            shallot.Logging.log_error(traceback.format_exc())
        Utils.clearcaches()


shallot.Actions.register(RunAniseTasksActionMenu, predicates=[shallot.Actions.HideOnSelectionLevelPredicate(),
                                                              shallot.Actions.PositionIndexPredicate(178300000)])
shallot.PanelDetails.PanelDetail.register(AnisePanelDetailProjectName())
