# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of Yaji.
#
# Yaji is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yaji is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

from typing import *

import threading


class DialogController:
    """
    A controller for dialogs.
    """

    def __init__(self, app: 'Application'):
        self.__app = app
        self.__dialogs = []
        self.__nextdialogid = 0
        self.__lock = threading.Lock()

    def show_dialog(self, dialog: 'Dialog'):
        """
        Shows a dialog.

        :param dialog: The dialog to show.
        """
        with self.__lock:
            if dialog.dialogid is not None:
                return
            # noinspection PyProtectedMember
            dialog._set_dialogid(self.__nextdialogid)
            self.__nextdialogid += 1
            self.__dialogs.append(dialog)
            self.__app.triggerevent("_yj_showdialog", dialog=dialog)

    # noinspection PyProtectedMember
    def close_dialog(self, dialog: 'Dialog'):
        """
        Closes a dialog.

        :param dialog: The dialog to close.
        """
        with self.__lock:
            if dialog.dialogid is None:
                return
            self.__dialogs.remove(dialog)
            self.__app.triggerevent("_yj_closedialog", dialogid=dialog.dialogid)
            dialog._exec_closed_handlers()
            dialog._set_dialogid(None)

    def get_dialogs(self) -> List['Dialog']:
        """
        Returns a list of all open dialogs.
        """
        return list(self.__dialogs)

    def get_dialog_by_id(self, dialogid: int) -> Optional['Dialog']:
        """
        Returns a dialog by id.

        :param dialogid: The dialog id.
        """
        with self.__lock:
            for dlg in self.__dialogs:
                if dlg.dialogid == dialogid:
                    return dlg


if TYPE_CHECKING:  # ugly hack for type annotations
    from .app import *
    from .gui import *
