/*
 @licstart  The following is the entire license notice for the JavaScript code in this page.

 Copyright (C) 2014-2021, Josef Hahn

 This file is part of Yaji.

 Yaji is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Yaji is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Yaji.  If not, see <http://www.gnu.org/licenses/>.

 @licend  The above is the entire license notice for the JavaScript code in this page.
*/

"use strict";

/**
* An i18n-aware (translatable) string.
*/
class TrStr {

    constructor(stringname, args) {
        this.stringname = stringname;
        this.args = args;
    }

    toString() {
        var result = clove.i18n[this.stringname];
        var replacements = [];
        for (var keyname in this.args) {
            var keystring = "{" + keyname + "}";
            var keyvalue = this.args[keyname];
            var cur = 0;
            while (cur < result.length) {
                var i = result.indexOf(keystring, cur);
                if (i == -1)
                    break;
                replacements.push([i, i+keystring.length, keyvalue]);
                cur = i + keystring.length;
            }
        }
        replacements.sort(function (a, b) { return b[0] - a[0]; });
        for(let replacement of replacements)
            result = result.slice(0, replacement[0]) + replacement[2] + result.slice(replacement[1])
        return result;
    }

    static _from_simple_repr_(data) {
        return new TrStr(data.stringname, data.args);
    }

}

/**
* A special clove::PopupMenuButton connected to application's action execution framework.
* |infrastructure|
*/
class YajiPopupMenuButton extends clove.PopupMenuButton {

    doinit() {
        super.doinit();
        this.OnActionTriggered.addHandler(yaji._runActionCallFromEvent);
    }

}

/**
* A special clove::AjaxAsyncDatasource backed by a datasource on application code side.
* See also app.Application.create_datastore() and app.Application.bindserver().
*/
class YajiDataStore extends clove.AjaxAsyncDatasource {

    constructor(dscfg) {
        super({
            data_pullUrl: "_yj_datastore_pull",
            data_pushUrl: "_yj_datastore_push",
            params: {name: dscfg.name},
            rootType: "data",
            ajaxAdditionalArgs: {dataType: 'json'},
        });
        this.name = dscfg.name;
        this._curtain = __EventBinding._curtainDefToCurtain(dscfg.curtain);
    }

    do_async_push(ptr, v, requestConfig, __call_super) {
        var self = this;
        if (!this._curtain || __call_super == '__call_super')
            return super.do_async_push(ptr, v, requestConfig);
        else
            this._curtain.execute(function(){ self.do_async_push(ptr, v, requestConfig, '__call_super'); });
    }

}

/**
* The application configuration.
* See also app.Application.appconfig().
*/
class YajiConfiguration {

    constructor(app) {
        this._handlers = {};
        this._appliedvalues = undefined;
        this._app = app;
        app.addEventHandler("_yj_appconfig_set", function(e){
            app.appconfig._applyValue(e.key, e.value);
        });
    }

    _applyValues(dict) {
        this._appliedvalues = {};
        for(var key in dict)
            if (dict.hasOwnProperty(key))
                this._appliedvalues[key] = dict[key];
    }

    _applyValue(key, value) {
        var hasappliedvalue = this._appliedvalues.hasOwnProperty(key);
        var oldappliedvalue = this._appliedvalues[key];
        var isnewvalue = (!hasappliedvalue || (oldappliedvalue !== value));
        this._appliedvalues[key] = value;
        var handlers = this._handlers[key];
        if (handlers)
            for (var ih=0;ih<handlers.length;ih++) {
                var hh= handlers[ih];
                if (isnewvalue || hh[1])
                    hh[0](value, oldappliedvalue, hasappliedvalue);
            }
    }

    /**
    * Adds a configuration handler for a config key. It gets called whenever the value for this key
      changes, and also directly if there already is something stored for that key at execution time.
    * \param key The configuration key.
    * \param fct The handler function.
    * \param runAlsoOnNeutralAssignments If to run it also for assignments that have not changed the value.
    */
    addHandler(key, fct, runAlsoOnNeutralAssignments) {
        if (this._appliedvalues === undefined)
            throw new Error("YajiConfiguration.addHandler called before yaji was ready.");
        if (!this._handlers.hasOwnProperty(key))
            this._handlers[key] = [];
        this._handlers[key].push([fct, runAlsoOnNeutralAssignments]);
        if (this._appliedvalues.hasOwnProperty(key)) {
            var value = this._appliedvalues[key];
            fct(value, undefined, false);
        }
    }

    /**
    * Returns a configuration value by key.
    * \param key The configuration key.
    */
    getConfig(key) {
        return this._appliedvalues[key];
    }

    /**
    * Requests to set a configuration value for a key.
    * Note: This happens asynchronously by a request to the backend.
    * \param key The configuration key.
    * \param value The new configuration value.
    */
    setConfig(key, value) {
        this._app.ajax({
            url: '_yj_setappconfigvalue',
            data: {key: key, value: JSON.stringify(value)},
        });
    }

}

/**
* Internal handling of client events. See YajiClient.addEventHandler().
* |infrastructure|
*/
class YajiClientEvents {

    constructor(app) {
        this._app = app;
        this._eventhandlers = {};
        this._eventidhandlers = {};
    }

    _addEventHandler(key, fct) {
        if (!this._eventhandlers.hasOwnProperty(key))
            this._eventhandlers[key] = [];
        this._eventhandlers[key].push(fct);
    }

    _processEvent(event) {
        for (var eventhandler of this._eventhandlers[event.name] ||[])
            eventhandler(event);
    }

    _getNextEvent() {
        var self = this;
        if (!this._app.isrunning)
            return;
        this._app.ajax({
            url: '_yj_pullevent',
            data: {lastid: self._app.eventlastid},
            method: 'GET',
            onsuccess: function(data) {
                for(var i=0; i<data.length; i++) {
                    var ddata = data[i];
                    self._app.eventlastid = ddata.id;
                    try {
                        self._processEvent(ddata.data);
                    }
                    catch (e) {
                        setTimeout(function(){throw e;});
                    }
                    var eventidhdlrs = self._eventidhandlers[ddata.id];
                    if (eventidhdlrs) {
                        delete self._eventidhandlers[ddata.id];
                        for (var eventidhdlr of eventidhdlrs)
                            eventidhdlr();
                    }
                }
                setTimeout(function(){ self._getNextEvent(); }, 0);
            },
            onerror: function() {
                self._app._isrunning = false;
                self._app.uiShutdown();
            }
        });
    }

    _waitForEventId(eventid) {
        var self = this;
        return {then: function(resolve, reject) {
            if (eventid <= self._app.eventlastid)
                resolve();
            else {
                var hdlrs = self._eventidhandlers[eventid];
                if (!hdlrs)
                    hdlrs = self._eventidhandlers[eventid] = [];
                hdlrs.push(function(){ resolve(); });
            }
        }};
    }

}

/**
* Abstract filesystem picker dialog builder. See subclasses.
*/
class AbstractFilesystemDialog {

    constructor(config) {
        this.config = clove.utils.applyDefaultsToConfig(config, {
            startPath: "/",
            title: clove.i18n.YajiFilesystem,
            headerWidget: {height: '1px'},
            headerText: "",
            showHidden: true,
            onlyDirs: false,
            onsuccess: function(){},
            oncancel: function(){},
            app: undefined,
        });
    }

    _listfsdata(path) {
        return {path: path, show_hidden: this.config.showHidden?"1":"0", only_dirs: this.config.onlyDirs?"1":"0"};
    }

    _showdir(path) {
        var self = this;
        this.dirlist.setDatasource();
        this.lblcurrdir.setLabel(clove.i18n.YajiCurrent + ": ...");
        this.btngoto.setEnabled(false);
        this.config.app.ajax({
            url: '_yj_listfs',
            data: this._listfsdata(path),
            onsuccess: function(data) {
                self.btngoto.setEnabled(true);
                self.currpath = data.path;
                self.lblcurrdir.setLabel(clove.i18n.YajiCurrent + ": " + data.path);
                var ds = clove.utils.arrayToDatasource(data.dirs);
                ds.insertColumn(0);
                for(var ir=0;ir<ds.rowCount();ir++) {
                    var ptr = ds.valuePointer(ir, 0);
                    ds.setValue(ptr, data.dirs[ir][1]);
                    ds.setMetadata(ptr, "icon", clove.Icon.bySymbol(data.dirs[ir][0] ? "⠀" : "📂"));
                }
                self.dirlist.setDatasource(ds);
            },
        });
    }

    /**
    * Shows the dialog and calls either `onsuccess` or `oncancel`.
    */
    showDialog() {
        var self = this;
        var dlgres = clove.Dialog.show({
            view: 'clove.Dialog',
            title: this.config.title,
            body: {
                rows: [
                    {label: this.config.headerText, styleClass: 'clovestyle_vmargin_tiny clovestyle_hmargin_small', width: '1px', hstretch: 1},
                    this.config.headerWidget,
                    {rows: [{label: '', name: 'lblcurrdir', styleClass: 'yaji_filebrowser_currentdir'}], styleClass: 'clovestyle_panel_bar'},
                    {view: 'ScrollView', height: '150pt', hstretch: 1, body:
                        {view: 'DataView', showOnlyFirstColumn: true, name: 'dirlist', allowSelection: true, hideExpanders: true, gridVisible: false, granularity: 'row', styleClass: 'clovestyle_margin_small'},
                    },
                    {cols: [
                        {view: 'Button', label: clove.i18n.YajiGoto, name: 'btngoto', styleClass: 'clovestyle_marginnotleft_small', icon: clove.Icon.bySymbol(clove.symbols.CheckMark)},
                        {view: 'Button', label: clove.i18n.YajiDrives, name: 'btnchangedrive', styleClass: 'clovestyle_marginnotleft_small'},
                        {view: 'Button', label: clove.i18n.YajiCancel, name: 'btncancel', styleClass: 'clovestyle_marginnotleft_small'},
                    ]},
                ],
                width: '330pt',
            },
        }, {
            modal: true,
        });
        this.dirlist = dlgres.namescope.getByName('dirlist');
        this.lblcurrdir = dlgres.namescope.getByName('lblcurrdir');
        this.btncancel = dlgres.namescope.getByName('btncancel');
        this.btngoto = dlgres.namescope.getByName('btngoto');
        this.btnchangedrive = dlgres.namescope.getByName('btnchangedrive');
        this.currpath = undefined;
        if (!this.config.app.appconfig.getConfig("_yj_abstractfilesystemdialog_showChangeDriveButton"))
            this.btnchangedrive.setVisibility(clove.InvisibleCollapsed);
        this.dirlist.OnNodeActivated.addHandler(function(e){
            var eds = e.node.datasource;
            var npath = self.currpath + (self.currpath=="/" ? "" : "/") + eds.getValue(e.node);
            var isfile = eds.getValue(eds.valuePointer(e.node.irow, 1))[0];
            if (isfile) {
                dlgres.widget.close();
                if (self.config.onsuccess)
                    self.config.onsuccess(npath);
            }
            else
                self._showdir(npath);
        });
        this.btncancel.OnClicked.addHandler(function(e){
            dlgres.widget.close();
            if (self.config.oncancel)
                self.config.oncancel();
        });
        this.btngoto.OnClicked.addHandler(function(e){
            dlgres.widget.close();
            if (self.config.onsuccess)
                self.config.onsuccess(self.currpath);
        });
        this.btnchangedrive.OnClicked.addHandler(function(e){
            self._showdir("");
        });
        this._showdir(self.config.startPath);
        return dlgres;
    }

}

/**
* Open file dialog builder.
*/
class OpenFileDialog extends AbstractFilesystemDialog{

    constructor(config) {
        super(clove.utils.applyDefaultsToConfig(config, {
            title: clove.i18n.YajiOpenFile,
        }));
    }

    _showdir(path) {
        this.btngoto.setVisibility(clove.InvisibleCollapsed);
        super._showdir(path);
    }

}

/**
* Open directory dialog builder.
*/
class OpenDirectoryDialog extends AbstractFilesystemDialog{

    constructor(config) {
        super(clove.utils.applyDefaultsToConfig(config, {
            title: clove.i18n.YajiOpenDir,
            onlyDirs: true,
        }));
    }

}

/**
* A dialog (either modal or non-modal).
* |infrastructure|
*/
class Dialog {

    constructor(dlgdata) {
        this._dlgdata = dlgdata;
        this._native = undefined;
    }

    get dialogid() {
        return this._dlgdata.dialogid;
    }

    _show() {
        var self = this;
        this._native = clove.Dialog.show(this._dlgdata.cfg, this._dlgdata.showcfg);
        this._native.widget.OnDestroyed.addHandler(()=>{
            yaji.ajax({ url: '_yj_dialogs_close', data: {dialogid: self.dialogid} });
        });
    }

    _close() {
        try {
            this._native.widget.remove();
        }
        catch(e) {}
    }

    static _from_simple_repr_(data) {
        return new Dialog(data);
    }

}

/**
* Manager for backend controlled dialogs.
*/
class YajiDialogs {

    constructor(app) {
        var self = this;
        this._dialogs = {};
        app.ready(function(){
            app.addEventHandler("_yj_showdialog", function(e){
                self._showDialog(e.dialog);
            });
            app.addEventHandler("_yj_closedialog", function(e){
                self._closeDialog(e.dialogid);
            });
            app.ajax({
                url: '_yj_dialogs_list',
                success: function(dlgs) {
                    for (var dlg of dlgs)
                        self._showDialog(dlg);
                }
            });
        });
    }

    _showDialog(dlg) {
        if (this._dialogs.hasOwnProperty(dlg.dialogid))
            return;
        dlg._show();
        this._dialogs[dlg.dialogid] = dlg;
    }

    _closeDialog(dialogid) {
        var dlg = this._dialogs[dialogid];
        if (!dlg)
            return;
        dlg._close();
        delete this._dialogs[dialogid];
    }

}

/**
* User feedback manager. This is used for simple dialogs like message boxes, input boxes, and so on.
*/
class YajiUserFeedback {

    constructor(app) {
        this._app = app;
        this._currentrequest = undefined;
        this._currentrequestdlg = undefined;
        this._handlers = {};
        var self = this;
        this.registerHandler("message", function(feedbackcfg, doanswer){
            var result = clove.messageDialog({message: feedbackcfg.message, buttons: feedbackcfg.buttons});
            result.addHandler(function(e){
                doanswer(e.button);
            });
            return result.dlgres.widget;
        });
        this.registerHandler("input", function(feedbackcfg, doanswer){
            var result = clove.inputDialog({question: feedbackcfg.question, defaultAnswer: feedbackcfg.defaultanswer});
            result.addHandler(function(e){
                doanswer(e.answer);
            });
            return result.dlgres.widget;
        });
        this.registerHandler("password", function(feedbackcfg, doanswer){
            var result = clove.inputDialog({question: feedbackcfg.question, defaultAnswer: feedbackcfg.defaultanswer,
                                            contentWidgetConfig: {view: 'PasswordEditBox'}});
            result.addHandler(function(e){
                doanswer(e.answer);
            });
            return result.dlgres.widget;
        });
        this.registerHandler("multilineinput", function(feedbackcfg, doanswer){
            var result = clove.inputDialog({question: feedbackcfg.question, defaultAnswer: feedbackcfg.defaultanswer,
                                    contentWidgetConfig: {view: 'MultilineEditBox', width: '450pt', height: '400pt',
                                            OnKeyDown: function(e) {
                                                if (e.keyCode == 9) {
                                                    var v = e.widget.text();
                                                    var cursorPos = $(e.widget.inputnode).prop('selectionStart');
                                                    e.widget.setText(v.substring(0, cursorPos) + "    " + v.substring(cursorPos));
                                                    $(e.widget.inputnode).prop('selectionStart', cursorPos + 4);
                                                    $(e.widget.inputnode).prop('selectionEnd', cursorPos + 4);
                                                    e.originalEvent.stopPropagation();
                                                    e.originalEvent.preventDefault();
                                                }
                                            }},
                                    withKeyboardControl: false});
            result.addHandler(function(e){
                doanswer(e.answer);
            });
            return result.dlgres.widget;
        });
        this.registerHandler("choice", function(feedbackcfg, doanswer){
            var dlgres;
            function _requestanswerer(v) {
                return function() {
                    doanswer(v);
                    dlgres.widget.close();
                };
            }
            var choicebuttons = [];
            for(var i=0; i<feedbackcfg.choices.length; i++)
                choicebuttons.push({view: 'Button', label: feedbackcfg.choices[i], OnClicked: _requestanswerer(i)});
            dlgres = clove.conversationDialog({buttons: [clove.i18n.Cancel], body: {rows: [{view: 'Label', label: feedbackcfg.question}, {rows: choicebuttons}]}});
            dlgres.OnProceed.addHandler(function(e){
                doanswer(undefined);
            });
            return dlgres.widget;
        });
        this.registerHandler("filesystem", function(feedbackcfg, doanswer){
            var clss = (feedbackcfg.fstype == "directory") ? OpenDirectoryDialog : OpenFileDialog;
            return (new clss({
                onsuccess: function(newpath) {
                    doanswer(newpath);
                },
                oncancel: function() {
                    doanswer(undefined);
                },
                headerText: feedbackcfg.question,
                startPath: feedbackcfg.startpath,
                app: app,
            })).showDialog().widget;
        });
        app.ready(function(){  // do that late enough so others have a chance to register handlers
            app.appconfig.addHandler("_yj_userfeedback_current", function(newv, oldv, existsv) {
                if (!self._currentrequest || !newv || self._currentrequest.reqid != newv.reqid) {
                    if (self._currentrequestdlg) {
                        self._currentrequestdlg.remove();
                        self._currentrequestdlg = undefined;
                    }
                    self._currentrequest = newv;
                    if (newv) {
                        self._currentrequestdlg = self.show(newv, function(answer) {
                            app.ajax({
                                url: '_yj_answer_userfeedback',
                                data: {reqid: newv.reqid, answer: JSON.stringify([answer])},
                                method: 'POST',
                            });
                        });
                    }
                }
            });
        });
    }

    _answerRequest(feedbackcfg, answer) {
        this._app.ajax({
            url: '_yj_answer_userfeedback',
            data: {reqid: feedbackcfg.reqid, answer: JSON.stringify([answer])},
            method: 'POST',
        });
    }

    /**
    * Shows a feedback dialog by feedback configuration and calls a callback for the answer.
    * \param feedbackcfg Feedback configuration, incl. `kind` and some kind-dependent parameters.
    * \param onanswered The callback for the dialog answer.
    */
    show(feedbackcfg, onanswered) {
        var r = this._handlers[feedbackcfg.kind](feedbackcfg, onanswered);
        if (!r)
            throw new Error("It is not allowed for a YajiUserFeedback handler to return nothing.");
        return r;
    }

    /**
    * Registers a handler for a custom kind of user feedback dialog.
    * \param kind The custom dialog kind name.
    * \param fct The custom handler function for this dialog kind.
    */
    registerHandler(kind, fct) {
        this._handlers[kind] = fct;
    }

}

/**
* Browser side support for PyHtmlView.
*/
class PyHtmlViewSupport {

    /**
    * |infrastructure|
    */
    constructor(app) {
        this._app = app;
    }

    /**
    * Used by PyHtmlView when the user tries to close the window.
    */
    _onwindowclose() {
        var self = this;
        if (!this._app.isrunning)
            return true;
        else if (this._app.appconfig.getConfig("_yj_stopImplicitlyWhenBrowserClosed")) {
            this._app.stop();
            return true;
        }
        else if (this._app.appconfig.getConfig("_yj_showBrowserClosedNotification")) {
            var dlg = clove.messageDialog({message: clove.i18n.YajiPyHtmlViewSupportConfirmClose,
                                           buttons: [clove.i18n.Yes, clove.i18n.No]});
            dlg.addHandler(function(e){
                if (e.button == 0) {
                    self._app.OnUiShutdown.addHandler(function(){
                        $.ajax({
                            url: '_yj_application_tryclosebrowser',
                            method: 'POST',
                            data: {checks: '0'},
                            dataType: 'json',
                        });
                    });
                    self._app.stop(true);
                }
            });
        }
    }

}

/**
* Backend errors in answering ajax requests.
*/
class YajiAjaxError extends Error {

    constructor(url, httpcode, body) {
        super(`Backend error in '${url}'.`);
        this.url = url;
        this.httpcode = httpcode;
        this.body = body;
    }

    static _byJQuery(url, x) {
        return new YajiAjaxError(url, x.status, x.responseText);
    }

}


/**
* Handle object for an ajax request (triggered by YajiClient.ajax()).
*/
class YajiAjaxRequest {

    constructor(app) {
        this._app = app;
        this._additionalFinishedHandlers = [];
    }

    static _openRequests() {
        return new Set(yaji._openAjaxRequests);
    }

    _finish() {
        for (var fct of this._additionalFinishedHandlers)
            fct();
    }

    async _waitFinished() {
        var self = this;
        return {then: function(resolve, reject) {
            if (self._app._openAjaxRequests.has(self))
                self._additionalFinishedHandlers.push(function(){ resolve(); });
            else
                resolve();
        }};
    }

}

/**
* User interface curtains are used for disabling parts of the user interface during some backend operations.
*
* The default implementation disables the entire user interface with a modal busy animation for that, so in many cases
* you have to subclass an own implementation.
*/
class Curtain {

    constructor() {
        this._cnt = 0;
        this._initialajaxreqs = undefined;
        this.__modalpanel = undefined;
    }

    /**
    * Implements closing the curtain, so blocking some parts of the user interface in some ways.
    * |dooverride_optional|
    * If you override it, also override open().
    */
    close() {
        this.__modalpanel = clove.utils.popup({view: 'Widget', busy: true}, {x: '0', y: '0', height: '100%', width: '100%'}).widget;
    }

    /**
    * Implements opening the curtain, so unblocking the user interface again.
    * |dooverride_optional|
    * If you override it, also override close().
    */
    open() {
        this.__modalpanel.remove();
    }

    /**
    * Executes a function inside the curtain, so the curtain gets closed before, and gets opened when processing
    * is idle again.
    */
    async execute(fct) {
        var self = this;
        if (!this._cnt++) {
            this._initialajaxreqs = YajiAjaxRequest._openRequests();
            this.close();
        }
        fct();
        if (!--this._cnt) {
            var idleInARow = 0;
            while (idleInARow < 3) {
                var newajaxreqs = [...YajiAjaxRequest._openRequests()].filter(x => !self._initialajaxreqs.has(x));
                if (newajaxreqs.length > 0) {
                    idleInARow = 0;
                    var lasteventid = await yaji.ajax({url: "_yj_lasteventid"});
                    await yaji._clientevents._waitForEventId(lasteventid);
                    for (var newajaxreq of newajaxreqs)
                        await newajaxreq._waitFinished();
                }
                else
                    idleInARow++;
            }
            this._initialajaxreqs = undefined;
            this.open();
        }
    }

}

class __DataBinding {

    static _from_simple_repr_(data) {
        data.datasource = yaji._getDatasource(data.datasource);
        data.dataDirection = data.data_direction;
        data.valueConverter = data.convertername ? new (eval(data.convertername))() : undefined;
        return clove.databind(data);
    }

}

class __EventBinding {

    static _from_simple_repr_(data) {
        function f(e) {
            if (data.function._yj_is_yjaction == 'dict')
                return data.function({__yj_uieventdata: e});
            else
                return data.function(e);
        }
        if (data.curtain) {
            var origf = f;
            var curtain = __EventBinding._curtainDefToCurtain(data.curtain);
            f = function() {
                var a = arguments;
                curtain.execute(()=>{origf(...a);});
            }
        }
        return f;
    }

    static _curtainDefToCurtain(curtdef) {
        if (curtdef === true)
            curtdef = new Curtain();
        if (typeof curtdef === "string" || curtdef instanceof String)
            curtdef = eval(curtdef);
        if (curtdef instanceof Function)
            curtdef = new curtdef();
        return curtdef;
    }

}

class __Datasource {

    static _from_simple_repr_(data) {
        return yaji._getDatasource(data);
    }

}

class __Date {

    static _from_simple_repr_(data) {
        return new Date(parseFloat(data.value)*1000);
    }

}

class __BackendFunction {

    static _from_simple_repr_(data) {
        function callfct(args) {
            var sargs = {};
            for (var k in args) {
                var v = args[k];
                if (!(typeof v === 'string' || v instanceof String))
                    v = yaji.tryJsonStringify(v);
                sargs[k] = v;
            }
            yaji.ajax({ url: data.url, method: 'POST', data: {...data.funcargs, ...sargs} });
        }
        callfct._yj_is_yjaction = 'dict';
        return callfct;
    }
}

class __BrowserFunction {

    static _from_simple_repr_(data) {
        function callfct() {
            eval(data.funcname)(...data.funcargs, ...arguments);
        }
        callfct._yj_is_yjaction = 'array';
        return callfct;
    }
}

class __Action {

    static _from_simple_repr_(data) {
        return clove.utils.applyDefaultsToConfig({name: data.function}, data);
    }
}

class __Submenu {

    static _from_simple_repr_(data) {
        return data;
    }
}

class __Separator {

    static _from_simple_repr_(data) {
        return clove.MenuSeparator;
    }
}

class __YajiDataStore_Node {

    static _from_simple_repr_(data) {
        return data;
    }

}

class __Icon {

    static _from_simple_repr_(data) {
        if (data.srcfunc)
            return clove.Icon[data.srcfunc](data.src);
    }

}

/**
* The application main object. Usually you can access it by just `yaji`.
*/
class YajiClient {

    constructor() {
        var self = this;
        clove._ajaxprovider = function(config) {
            if (config.success) {
                var isuccessfct = config.success;
                config.success = function(data) {
                    self._parseajaxdata(data);
                    isuccessfct(data);
                }
            }
            $.ajax(config);
        };
        this._openAjaxRequests = new Set();
        XMLHttpRequest.prototype._bef_yj_open = XMLHttpRequest.prototype.open;
        XMLHttpRequest.prototype.open = function(_, ourl) {
            if (ourl.indexOf("://") == -1 && !ourl.startsWith("_yj_pullevent?")) {
                this._yj_req = new YajiAjaxRequest(self);
                self._openAjaxRequests.add(this._yj_req);
                this.addEventListener("readystatechange", function(){
                    if (this.readyState >= XMLHttpRequest.DONE) {
                        self._openAjaxRequests.delete(this._yj_req);
                        this._yj_req._finish();
                    }
                });
            }
            this._bef_yj_open(...arguments);
        };
        this._mainview = undefined;
        this._datasources = {};
        this._head1 = undefined;
        this._head2 = undefined;
        this._uiShutdownTriggered = false;
        this._isready = false;
        this._isrunning = false;
        this._readyHandlers = [];
        this._ajaxParsersCache = {};
        this._clientevents = new YajiClientEvents(this);
        this._appconfig = new YajiConfiguration(this);
        this._dialogs = new YajiDialogs(this);
        this._userfeedback = new YajiUserFeedback(this);
        this._pyhtmlview = new PyHtmlViewSupport(this);
        this._hasExternalCloseControls = undefined;
        this._OnUiShutdown = new clove.Event();
        this._OnUiShutdown.addHandler(function(){ self._showShutdownDialog(); });
        this.addEventHandler("_yj_stopui", function(event){
            self._isrunning = false;
            self.uiShutdown();
            if (event.targeturl)
                location.href = event.targeturl;
        });
        this.addEventHandler("_yj_showunhandlederror", function(event){
            self._errorPopup(event.errortext);
        });
        this.addEventHandler("_yj_datastore_update", async function(event){
            var ds = await self.getDatastore(event.datasource);
            var objids = event.id;
            var bo = ds._rootnode;
            for (var i=0; i<objids.length; i++) {
                var objid = objids[i];
                var rcell;
                for (var iro=0; iro<bo.childs.length; iro++) {
                    var row = bo.childs[iro];
                    for (var ice=0; ice<row.length; ice++) {
                        var cell = row[ice];
                        if (cell && cell.ajaxid == objid) {
                            rcell = cell;
                            break;
                        }
                    }
                    if (rcell)
                        break;
                }
                if (!rcell)
                    return;
                else
                    bo = rcell;
            }
            ds.async_pull(bo.valuePointer(), {reason: 'YajiDataStore', sub: 'update'});
        });
    }

    /**
    * Instance of YajiConfiguration.
    */
    get appconfig() {
        return this._appconfig;
    }

    /**
    * Instance of YajiUserFeedback.
    */
    get userfeedback() {
        return this._userfeedback;
    }

    /**
    * Instance of PyHtmlViewSupport.
    */
    get pyhtmlview() {
        return this._pyhtmlview;
    }

    /**
    * Instance of clove::MainView.
    */
    get mainview() {
        return this._mainview;
    }

    /**
    * If the application is currently running.
    */
    get isrunning() {
        return this._isrunning;
    }

    /**
    * If this client view has dedicated external close controls (e.g. via PyHtmlView).
    */
    get hasExternalCloseControls() {
        return this._hasExternalCloseControls;
    }

    /**
    * If the application is static. See app.Application.storeasstaticapplication().
    */
    get isStaticApplication() {
        if (!this._isready)
            throw new Error("Must not be called before the application is ready.");
        return this.appconfig.getConfig("_yj_isStaticApplication");
    }

    /**
    * A clove::Event that is triggered on user interface shutdown. See also YajiClient.uiShutdown().
    */
    get OnUiShutdown() {
        return this._OnUiShutdown;
    }

    /**
    * Adds a handler function for a backend event. See app.Application.triggerevent().
    * \param key The event key (or event 'name').
    * \param fct The handler function. Receives the event data as first argument.
    */
    addEventHandler(key, fct) {
        this._clientevents._addEventHandler(key, fct);
    }

    /**
    * Returns a json representation for an object, just containing primitive values, arrays and dicts.
    * \param o The object to serialize.
    */
    tryJsonStringify(o) {
        return JSON.stringify(o, (name, val) => {
            var tval = typeof val;
            if (tval != 'object' && tval != 'function')
                return val;
            var pval = Reflect.getPrototypeOf(val);
            if (pval == Object.prototype || pval == Array.prototype || pval == clove.EventArgs.prototype)
                return val;
        }) ||'undefined';
    }

    /**
    * Completely stops the application, including application code side.
    * See app.Application.stop() for explanation about stop procedure, and also YajiClient.uiShutdown().
    * \param forceBlockUiDuringRequest If to force freezing the user interface by a modal panel (which automatically
    *                                  happens only in some situations) during stop request.
    */
    stop(forceBlockUiDuringRequest) {
        var self = this;
        if (!this.appconfig.getConfig("_yj_skipShutdownDialog") || forceBlockUiDuringRequest)
            var modalpanel = clove.utils.popup({view: 'Widget'}, {modal: true, fullyTransparent: false}).widget;
        else
            var modalpanel = undefined;
        $.ajax({
            url: '_yj_application_stop',
            method: 'POST',
            dataType: 'json',
            success: function() {
                if (modalpanel)
                    modalpanel.remove();
                self.uiShutdown();
            }
        });
    }

    /**
    * Makes an ajax request.
    *
    * It returns a Promise-like object, so it can be `await`ed.
    *
    * `config` may contain the same as for `$.ajax`, and also:
    * - `finished`: Callback that gets executed after either success or error.
    * - `alsoAfterShutdown`: If to make that request even if the ui is already stopped.
    * \param config The request configuration.
    */
    ajax(config) {
        var self = this;
        var configurl;
        config = clove.utils.applyDefaultsToConfig(config, {
            url: 'get',
            method: 'GET',
            dataType: 'json',
            onfinished: config.finished || function(){},
            onsuccess: config.success || function(){},
            onerror: config.error || function(x, t, e) { throw YajiAjaxError._byJQuery(configurl, x); },
            alsoAfterShutdown: false,
        });
        configurl = config.url;
        if (!config.alsoAfterShutdown && !this.isrunning)
            return;
        var _onsuccess, _onerror;
        config.success = function(data) {
            if (!config.alsoAfterShutdown && !self.isrunning)
                return;
            if (_onsuccess)
                return _onsuccess(data);
            if (config.onsuccess)
                config.onsuccess(data);
            if (config.onfinished)
                config.onfinished(data);
        };
        config.error = function(x, t, e) {
            if (!config.alsoAfterShutdown && !self.isrunning)
                return;
            if (_onerror)
                return _onerror(YajiAjaxError._byJQuery(configurl, x));
            if (config.onerror)
                config.onerror(x, t, e);
            if (config.onfinished)
                config.onfinished(undefined, x, t, e);
        };
        clove._ajaxprovider(config);
        var _thenres;
        return {then: function(resolve, reject) {
            if (_thenres)
                (_thenres[1]) ? reject(_thenres[1][0]) : resolve(_thenres[0]);
            else {
                _onsuccess = (d) => { _thenres=[d]; resolve(d); }
                _onerror = (e) => { _thenres=[undefined, [e]]; reject(e); }
            }
        }};
    }

    /**
    * Returns a application code side datasource by name.
    * See also app.Application.create_datastore() and app.Application.binddata().
    * \param name The datasource name.
    */
    async getDatastore(name) {
        if (!this._datasources.hasOwnProperty(name))
            this._getDatasource(await this.ajax({ url: '_yj_datastore_info', config: {name: name} }));
        return this._datasources[name];
    }

    /**
    * Returns a browser side datasource by name.
    * See also app.Application.create_clientlocal_datastore() and app.Application.binddata().
    * \param name The datasource name.
    */
    getClientlocalDatastore(name) {
        return this._getDatasource({name: name, islocal: true});
    }

    /**
    * Shuts down the user interface (without killing the backend).
    * See also YajiClient.stop().
    */
    uiShutdown() {
        if (this._uiShutdownTriggered)
            return;
        this._uiShutdownTriggered = true;
        this.OnUiShutdown.trigger();
    }

    /**
    * Executes a function on startup once the yaji app is initialized.
    * \param fct The function to execute.
    */
    ready(fct) {
        if (this._isready)
            setTimeout(fct);
        else
            this._readyHandlers.push(fct);
    }

    _initialize(isrunning, reopened, eventlastid, appconfigs, i18n) {
        var self = this;
        this._isrunning = isrunning;
        this.eventlastid = eventlastid;
        for (var trk in i18n)
            clove.i18n.addString(trk, i18n[trk]);
        $(document).ready(function(){
            self._hasExternalCloseControls = Boolean(window.pyhtmlview);
            clove.populateUI(function(){
                if (isrunning) {
                    clove.build({view: 'MainView', name: 'mainview'});
                    self._mainview = clove.getByName('mainview');
                    self._parseajaxdata(appconfigs);
                    self.appconfig._applyValues(appconfigs);
                    if (!self.appconfig.getConfig("_yj_isStaticApplication")) {
                        self._clientevents._getNextEvent();
                        window.onerror = function(e, src, line) {
                            self.ajax({
                                url: '_yj_unhandled_client_error',
                                data: { error: String(e), src: String(src||""), line: String(line||"") },
                                method: 'POST',
                                onsuccess: function(data) {
                                    if (data && data.shutdown)
                                        self.stop();
                                },
                                onerror: function() {}
                            });
                        };

                    }
                    self.appconfig.addHandler("_yj_takeoverurl", function(s) {
                        if (s)
                            location.href = s;
                    });
                    self.appconfig.addHandler("_yj_head1", function(s) {
                        self._head1 = s;
                        self._mainview.setHead1(s);
                        if (!self.appconfig.getConfig("skipsetwindowtitle"))
                            self._refreshWindowTitle();
                    });
                    self.appconfig.addHandler("_yj_head2", function(s) {
                        self._head2 = s;
                        self._mainview.setHead2(s);
                        if (!self.appconfig.getConfig("skipsetwindowtitle"))
                            self._refreshWindowTitle();
                    });
                    self.appconfig.addHandler("_yj_bodyleft", function(cfg) {
                        self._mainview.setBodyLeft(cfg);
                    });
                    self.appconfig.addHandler("_yj_bodyright", function(cfg) {
                        self._mainview.setBodyRight(cfg);
                    });
                    self.appconfig.addHandler("_yj_switchtobody", function(cfg) {
                        if (cfg === 0)
                            self._mainview.switchToLeftBody();
                        else if (cfg === 1)
                            self._mainview.switchToRightBody();
                    });
                    self.appconfig.addHandler("_yj_actions", function(actions) {
                        self._mainview.setActions(actions);
                    });
                    self.appconfig.addHandler("_yj_headcontrol", function(cfg) {
                        self._mainview.setHeadControl(cfg);
                    });
                    self.appconfig.addHandler("_yj_sidebar", function(cfg) {
                        self._mainview.setSidebar(cfg);
                    });
                    self.appconfig.addHandler("_yj_splitterposition", function(v) {
                        self._mainview.setSplitterPosition(v);
                    });
                    self.appconfig.addHandler("_yj_bodyleftviewactionlabel", function(label) {
                        self._mainview.setBodyLeftViewActionLabel(label);
                    });
                    self.appconfig.addHandler("_yj_bodyrightviewactionlabel", function(label) {
                        self._mainview.setBodyRightViewActionLabel(label);
                    });
                    self.appconfig.addHandler("_yj_showonly", function(v) {
                        self._mainview.setShowOnly(v);
                    });
                    self._mainview.OnActionTriggered.addHandler(self._runActionCallFromEvent);
                    self.appconfig.addHandler("_yj_icon", function (){ self._refreshMainviewIcon(); });
                    self.appconfig.addHandler("_yj_mainviewIcon", function (){ self._refreshMainviewIcon(); });
                    self._isready = true;
                    for (var i=0; i<self._readyHandlers.length; i++)
                        self._readyHandlers[i]();
                    self._readyHandlers = undefined;
                    if (self.appconfig.getConfig('_yj_showBrowserClosedNotification')) {
                        if (reopened) {
                            var notif = clove.notifications.notify({
                                rows: [
                                    {label: clove.i18n.YajiClosedNotificationText},
                                    {cols: [{view: 'Button', label: clove.i18n.YajiClosedNotificationButtonLabel, name: 'btnclose'}]},
                                ]},
                            {
                                timeout: 30,
                                closeable: true,
                            });
                            notif.getByName('btnclose').OnClicked.addHandler(function(){
                                self.stop();
                            });
                        }
                    }
                }
                else
                    self.uiShutdown();
            });
        });
    }

    _parseajaxdata (data, _p, _n) {
        if (data && typeof data !== "string") {
            for (var foo in data)
                this._parseajaxdata(data[foo], data, foo);
            var soc = data._yj_ser_obj_;
            if (soc) {
                delete data._yj_ser_obj_;
                var ap = this._ajaxParsersCache[soc];
                if (!ap)
                    ap = this._ajaxParsersCache[soc] = eval(soc+"._from_simple_repr_");
                _p[_n] = ap(data);
            }
        }
    }

    _refreshWindowTitle() {
        var title = (this._head1!==undefined) ? this._head1 : "";
        if (typeof this._head2 === 'string' || this._head2 instanceof String)
            title += ": " + this._head2;
        document.title = title;
    }

    _refreshMainviewIcon() {
        this._mainview.setIcon(this.appconfig.getConfig('_yj_mainviewIcon') || this.appconfig.getConfig('_yj_icon'));
    }

    _runActionCallFromEvent(e) {
        if (e.action && e.action._yj_is_yjaction)
            e.action();
    }

    _getDatasource(dscfg) {
        if (!this._datasources.hasOwnProperty(dscfg.name)) {
            var ds = dscfg.islocal ? new clove.NativeDatasource({readonly: false}) : new YajiDataStore(dscfg);
            this._datasources[dscfg.name] = ds;
            if (dscfg.init_with) {
                function _addCell(cfg, ptr) {
                    ds.setValue(ptr, cfg.value);
                    for (var ic=0; ic<cfg.colcount; ic++)
                        ds.appendColumn(ptr);
                    for (var ir=0; ir<cfg.rowcount; ir++) {
                        ds.appendRow(ptr);
                        for (var ic=0; ic<cfg.colcount; ic++)
                            _addCell(cfg.childs[ir][ic], ds.valuePointer(ir, ic, ptr));
                    }
                }
                _addCell(dscfg.init_with, undefined);
            }
        }
        return this._datasources[dscfg.name];
    }

    _errorPopup(errortext) {
        clove.notifications.notify({
            rows: [
                {label: clove.i18n.YajiProgramErrorOccurred},
                {
                    view: 'Expander', expandedLabel: clove.i18n.YajiHideDetails,
                    collapsedLabel: clove.i18n.YajiShowDetails,
                    body: { label: errortext, styleClass: 'yaji_errornotification_details', hstretch: 1, },
                },
            ],
        }, { styleClass: 'yaji_errornotification', closeable: true, });
    }

    _showShutdownDialog(){
        if (!this.appconfig.getConfig("_yj_skipShutdownDialog")) {
            clove.Dialog.show({
                view: 'Dialog',
                title: this.appconfig.getConfig("_yj_head1"),
                body: {cols: [
                    {view: 'IconView', icon: clove.Icon.bySymbol(clove.symbols.Warning), styleClass: 'clovestyle_margin_medium'},
                    {label: clove.i18n.YajiIsShutDown, styleClass: 'clovestyle_margin_medium'},
                ]}
            }, { modal: true });
        }
    }

}

window.yaji = new YajiClient();
yaji.Curtain = Curtain;
yaji.TrStr = TrStr;
yaji.OpenFileDialog = OpenFileDialog;
yaji.OpenDirectoryDialog = OpenDirectoryDialog;
