# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for building with gnu autotools.
"""

from typing import *
import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

buildmake = anise.framework.features.loadfeature("build.make").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION, requires=["milieus"])
    def _initproject():
        universe.milieus.register("autotools", AutotoolsMilieuComposite)


class AutotoolsMilieuComposite(buildmake.MakeMilieuComposite):
    """
    Build environment composite for autotools builds.

    See milieus.MilieuComposite for general infos.
    """

    def __init__(self, autoreconfpath: str = "autoreconf", makepath: str = "make", processcount: int = 1):
        """
        :param autoreconfpath: The path to the :samp:``autoreconf` tool.
        :param makepath: The path to the :samp:``make` tool.
        :param processcount: The maximum number of concurrent processes to execute.
        """
        super().__init__(makepath, processcount)
        self.autoreconfpath = autoreconfpath
        self.autoreconfcmd = ["{autoreconfpath}", "[[args]]"]

    def call_autoreconf(self, *args, raise_on_errors="unable to run autoreconf", **kwargs):
        cmd = [x.format(**self.__dict__) for x in self.autoreconfcmd]
        i = cmd.index("[[args]]")
        cmd = cmd[:i] + list(args) + cmd[i + 1:]
        return self.call(*cmd, raise_on_errors=raise_on_errors, **kwargs)


def build(source: anise.framework.files.Filestructure, milieu: Optional['milieus.Milieu'] = None,
          autoreconf_args: Optional[List[str]] = None, configure_args: Optional[List[str]] = None,
          make_args: Optional[List[str]] = None, makeinstall_args: Optional[List[str]] = None):
    """
    Builds with autotools (and autoreconf).

    :param source: The file structure to build.
    :param milieu: The milieu to use.
    :param autoreconf_args: Additional arguments for the :samp:`autoreconf` call.
    :param configure_args: Additional arguments for the :samp:`./configure` call.
    :param make_args: Additional arguments for the :samp:`make` call.
    :param makeinstall_args: Additional arguments for the :samp:`make install` call.
    """
    with universe.executionscope.newsubscope("Building via autotools"):
        milieu = universe.milieus.get(milieu, defaultmilieu="autotools")
        sourcepath = milieu.createtempdir().path
        source.dl(sourcepath)
        result = milieu.createtempdir()
        with anise.utils.basic.ChDir(sourcepath):
            milieu.call_autoreconf("--install", *(autoreconf_args or []))
            milieu.call("./configure", *(configure_args or []), raise_on_errors="unable to run ./configure")
            milieu.call_make(*(make_args or []), raise_on_errors="unable to run make")
            milieu.call_make("DESTDIR="+result.path, *(makeinstall_args or []), "install",
                             raise_on_errors="unable to run 'make install'")
        return result


if TYPE_CHECKING:
    import anise.features.milieus as milieus
