# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for building with make.
"""

from typing import *
import multiprocessing
import os
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

milieus = anise.framework.features.loadfeature("milieus").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION, requires=["milieus"])
    def _initproject():
        universe.milieus.register("make", MakeMilieuComposite)


class MakeMilieuComposite(milieus.MilieuComposite):
    """
    Build environment composite for make builds.
    
    See :py:class:`milieus.MilieuComposite` for general infos.
    """

    def __init__(self, makepath: str = "make", processcount: Optional[int] = None):
        """
        :param makepath: The path to the `make` tool.
        :param processcount: The maximum number of concurrent processes to execute.
        """
        super().__init__()
        self.makepath = makepath
        self.processcount = processcount or round(multiprocessing.cpu_count() * 1.6)
        self.makecmd = ["{makepath}", "-j{processcount}", "[[args]]"]

    def call_make(self, *args, **kwargs):
        cmd = [x.format(**self.__dict__) for x in self.makecmd]
        i = cmd.index("[[args]]")
        cmd = cmd[:i] + list(args) + cmd[i+1:]
        return self.call(*cmd, **kwargs)

    def make(self, targetname=None, makefile=None):
        makecall = []
        if makefile:
            makecall += ["-f", makefile]
        if targetname:
            makecall.append(targetname)
        r, o = self.call_make(*makecall)
        if r != 0:
            raise anise.framework.exceptions.ProcessExecutionFailedError("unable to run make", o)


def build(source: anise.framework.files.Filestructure, milieu: 'milieus.Milieu' = None,
          targetname: Optional[str] = None, targetfile: Optional[str] = None) -> anise.framework.files.Filestructure:
    """
    Builds with make.

    :param source: The source file structure to build.
    :param milieu: The milieu to use.
    :param targetname: The name of the make target for creating.
    :param targetfile: The file name of the target file to create.
    """
    with universe.executionscope.newsubscope(f"Building '{targetname}' via make" if targetname
                                             else "Building via make"):
        milieu = universe.milieus.get(milieu, defaultmilieu="make")
        sourcepath = milieu.createtempdir().path
        source.dl(sourcepath)
        with anise.utils.basic.ChDir(sourcepath):
            milieu.make(targetname)
            if targetfile is None:
                targetfile = targetname
            targetpath = milieu.createtempdir().path
            result = anise.framework.files.Filestructure(targetpath + "/" + os.path.basename(targetfile))
            if os.path.isdir(targetfile):
                anise.framework.files.Filestructure(targetfile).dl(result.path)
            else:
                shutil.copy2(targetfile, targetpath)
            return result


if TYPE_CHECKING:
    import anise.features.milieus as milieus
