# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for specifying and processing third-party packages that are required by the project in some way.
"""

from typing import *
import anise.framework.features

from anise.framework.projects import universe


class Internals:

    class Pool:
        """
        Storage of all dependencies.
        """

        def __init__(self):
            self.list = []

        def add(self, dependency: 'Dependency') -> None:
            """
            Adds a dependency.

            :param dependency: The dependency to add.
            """
            self.list.append(dependency)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject() -> None:
        universe.dependencies.pool = Internals.Pool()
        universe.dependencies.getdependencydocument = Internals.getdependencydocument

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_homepage_section() -> None:
        def my_homepage_section(**_):
            return universe.dependencies.getdependencydocument()
        universe.homepage.sections.add(5000, "dependencies", "Dependencies", my_homepage_section)

    @staticmethod
    def getdependencydocument() -> str:
        """
        Creates a documentation source that describes the project dependencies in a human-readable form.

        It is mostly used for including it into Doxygen documentation source.

        :return: The documentation source.
        """
        depslist = [x for x in universe.dependencies.pool.list if x.visible]
        if len(depslist) == 0:
            dependencies = "There are no dependencies.\n"
        elif len(depslist) == 1:
            dependencies = (f"There is an external part that is used by {universe.name}. "
                            f"Many thanks to the project and all participants.\n")
        else:
            dependencies = (f"There are external parts that are used by {universe.name}. "
                            f"Many thanks to the projects and all participants. "
                            f"Some installation methods might handle dependency installation automatically, while "
                            f"others leave that up to you.\n")
        for dependency in depslist:
            icon = dependency.icon or "nullicon"
            comment = dependency.comment
            obj = dependency.displayname
            dtype = dependency.type
            dlm = ": " if (len(comment) > 0) else ""
            obj = obj.replace("`", "'")
            dependencies = f".. |icon_{icon}| image:: aniseicons/{icon}.png\n\n" + dependencies
            dependencies += f"\n|icon_{icon}| *{obj}*, {dtype} {dlm}{comment}\n\n"
        return dependencies


class Type:
    """
    Enumeration of different types of dependencies. It describes how strongly required a dependency is.
    """
    Included = "included"
    Required = "required"
    Required_HasAlternatives = "required (has alternatives)"
    Recommended_HasAlternatives = "recommended"
    Optional = "optional"


class Dependency:
    """
    A dependency.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, type: Type, objectname: Optional[str] = None, comment: str = "", icon: Optional[str] = None,
                 visible: bool = True, displayname: Optional[str] = None, **kwargs):
        """
        :param type: The dependency type.
        :param objectname: A descriptive name for this dependency (e.g. a library name).
        :param comment: A comment.
        :param icon: The name of a dependency icon.
        :param visible: If it is visible in the dependency list.
        :param displayname: A display variant of the object name.
        :param kwargs: Additional arguments.
        """
        self.deptype = self.type = type  # we should get rid of `type` as variable name at some time (old code uses it)
        self.objectname = objectname
        self.comment = comment
        self.icon = icon
        self.visible = visible
        self.displayname = displayname or objectname
        for k in kwargs:
            setattr(self, k, kwargs[k])
