# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for specifying and processing project documentation.
"""

from typing import *
import os

import anise.features.docrender
import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.framework.report
import anise.utils

from anise.framework.projects import universe


#: reStructuredText documentation piece.
RstTextsPiece = anise.features.docrender.RstTextsPiece

#: Python 3 support for api references.
Python3ApiReferenceLanguage = anise.features.docrender.Python3ApiReferenceLanguage

#: C++ support for api references.
CppApiReferenceLanguage = anise.features.docrender.CppApiReferenceLanguage

#: JavaScript support for api references.
JavaScriptApiReferenceLanguage = anise.features.docrender.JavaScriptApiReferenceLanguage

#: Doxygen comment-based described language support for api references.
DoxygenDescribedApiReferenceLanguage = anise.features.docrender.AbstractDoxygenSupportedApiReferenceLanguage


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.documentation.exports = Internals.Exports()
        universe.documentation.pool = Internals.Pool()
        universe.documentation.pieces = Internals.Pieces()
        universe.documentation.htmltheme = None
        universe.documentation.generate_readme = True
        universe.documentation.readme_params = {}
        universe.documentation.readme_params_ = {'heading': "Manual", 'include_in_toc': ['FULLREADME']}

        def _writedocumentations(destdir):
            with universe.executionscope.newsubscope("Adding documentation"):
                Internals.writedocumentations(destdir=destdir, throw_on_existing=True, fortype=PackageExport)
        universe.addhook(anise.framework.features.loadfeature("packages").featuremodule.HOOK_ENRICHRAWPACKAGE,
                         _writedocumentations)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION,
                                          provides=["_compose_readmes"], requires=["base"])
    def _compose_readmes():

        if not universe.documentation.generate_readme:
            return

        def token_in_map(t, m):
            for mk, ms in m:
                if mk == t:
                    return ms
            return None

        maturity = universe.base.getprojectstatusdescription()
        smaturity = (universe.docgen.heading('Maturity') + maturity) if maturity else ""
        readmedefaultloc = f"{universe.specialpaths.projectdir}/_meta/readme.rst"
        if os.path.isfile(readmedefaultloc):
            readmetxt = anise.utils.basic.readfromfile(readmedefaultloc, onestring=True)

            universe.documentation.exports.add(HomepageExport(key="FULLREADME",
                                                              **{**universe.documentation.readme_params_,
                                                                 **universe.documentation.readme_params}))
        else:
            readmetxt = ""
        if not token_in_map("FULLREADME", universe.documentation.pool.list):
            _lauthors = getattr(universe.base, "authors", [])
            _authors = " and ".join(([", ".join(_lauthors[0:-1])] if len(_lauthors) > 1 else []) + _lauthors[-1:])
            if getattr(universe.base, "longdescription", ""):
                universe.documentation.pool.add("_LONGDESCRIPTION", universe.base.longdescription)
                _slongdesc = (universe.docgen.heading('About') + ".. include:: _LONGDESCRIPTION")
            else:
                _slongdesc = ""

            def fulltxt():
                _sversion = getattr(universe.packages, "getversion", lambda: None)()
                if _sversion:
                    _sversion = f"You are currently reading the manual for version {_sversion}."
                else:
                    _sversion = ""
                if universe.base.license:
                    _slicense = f" under the terms of the {universe.base.license.name}"
                else:
                    _slicense = ""
                return f"""
{universe.docgen.heading('User Manual', variant='^', sub=True)}

{universe.docgen.heading('License')}

{universe.name} is written by {_authors}{_slicense}.

Please read the `LICENSE` file from the package and the :ref:`Dependencies<hhdependencies>` section for included
third-party stuff.

{_slongdesc}

{universe.docgen.heading('Up-to-date?')}

Are you currently reading from another source than the homepage? Are you in doubt if that place
is up-to-date? If yes, you should visit {universe.base.websiteurl or 'the homepage'} and check that. {_sversion}

{smaturity}

.. _hhdependencies:

{universe.docgen.heading('Dependencies')}

{universe.dependencies.getdependencydocument()}

{readmetxt}
                    """
            universe.documentation.pool.add("FULLREADME", fulltxt)
        if not os.path.exists(f"{universe.specialpaths.projectdir}/README.pdf"):
            universe.documentation.exports.add(PackageExport(outputfile="/README.pdf", format=ExportFormat.PDF,
                                                             **{**universe.documentation.readme_params_,
                                                                **universe.documentation.readme_params}))

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION, requires=["_compose_readmes"])
    def _add_homepage_section():

        def my_homepage_section(homepage, **_):
            documentation = "The following documentation is available:\n\n"
            hpexports = [x for x in universe.documentation.exports.list if type(x) is HomepageExport]
            hpexports = sorted(hpexports, key=lambda x: x.index)
            for iechan, echan in enumerate(hpexports):
                echan.generate(f"{homepage.path}/{echan.key}")
                documentation += f"`{echan.heading} <{echan.key}/{echan.key}.html>`_\n\n"
            if len(hpexports) > 0:
                return documentation

        universe.homepage.sections.add(3000, "documentation", "Documentation", my_homepage_section)

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_release_task_for_export_to_source():
        if len([x for x in universe.documentation.exports.list if isinstance(x, SourceExport)]) > 0:
            def create_documentations_in_source():
                Internals.writedocumentations(universe.specialpaths.projectdir, throw_on_existing=False,
                                              fortype=SourceExport)
            universe.releasing.tasks.add(create_documentations_in_source)

    class Pool:
        """
        Storage of all documentation text generators.
        """

        def __init__(self):
            self.list = []

        def add(self, key: str, src: Union[str, Callable[[], str]]) -> None:
            """
            Adds a documentation.

            :param key: The documentation key for storage (later used for referencing this text).
            :param src: The documentation source text (or a function that returns one).
            """
            self.list.append([key, src])

        def gettexts(self) -> List[Tuple[str, str]]:
            """
            Returns all texts.

            :return: A list of key/text tuples.
            """
            return [(x, (y() if callable(y) else y)) for x, y in self.list]

    class Exports:
        """
        Storage of all documentation export definitions (i.e. where and how documentation products are to be created).
        """

        def __init__(self):
            self.list = []

        def getbyname(self, name: str) -> List['Internals.Export']:
            """
            Returns all export definitions with a given name.

            :param name: A export definition name.
            :return: List of export definitions.
            """
            return [x for x in self.list if x.name == name]

        def add(self, export: 'Internals.Export') -> None:
            """
            Exports a documentation to some targets.

            :param export: A export definition that specifies a documentation export target.
            """
            self.list.append(export)

    class Pieces:
        """
        Storage for documentation pieces.
        """

        def __init__(self):
            self.__storage = []

        def add_api_reference(self, *, language: anise.features.docrender.AbstractApiReferenceLanguage,
                              name: Optional[str] = None, rootpath: Optional[str] = None, **b) -> None:
            """
            Adds an api reference piece to the storage of documentation pieces.

            :param language: The programming language that is associated with this api.
            :param name: The internal piece name.
            :param rootpath: The root path of the source code.
            :param b: Additional piece arguments. See :py:class:`docrender.ApiReferencePiece`.
            """
            rootpath = rootpath or ""
            self.__storage.append(anise.features.docrender.ApiReferencePiece(
                name or (rootpath.replace('/', '_') + "__default"), rootpath, language, **b))

        def get_pieces(self) -> List[anise.features.docrender.AbstractPiece]:
            """
            Returns a list of all documentation pieces.
            """
            return list(self.__storage)

    @staticmethod
    def writedocumentations(destdir: str, throw_on_existing: bool,
                            fortype: Type['Internals.AbstractFileExport']) -> None:
        """
        Immediately exports the documentation as the export definitions currently specify.

        :param destdir: The destination directory path.
        :param throw_on_existing: If to raise an exception for an already existing destination.
        :param fortype: Which export type to process.
        """
        doclist = [x for x in universe.documentation.exports.list if isinstance(x, fortype)]
        scopelist = [universe.executionscope.newsubscope(
            f"Generating {y.format} documentation " + (f"'{y.heading}'" if y.heading else y.name)) for y in doclist]
        for i, echan in enumerate(doclist):
            topscope = anise.framework.report.topscope()
            topscope.progress = anise.framework.report.ExecutionScope.ProgressDeterminate(float(i)/len(doclist))
            with scopelist[i]:
                fullfn = f"{destdir}/{echan.outputfile}"
                if throw_on_existing and os.path.exists(fullfn):
                    raise anise.framework.exceptions.BadFormatError(
                        "there already is a '{fullfn}', which is forbidden.".format(fullfn=fullfn))
                anise.utils.basic.mkdirs(os.path.dirname(fullfn))
                echan.generate(fullfn)

    class Export:
        """
        Defines a documentation export.
        """

        # noinspection PyShadowingBuiltins
        def __init__(self, format: str, name: Optional[str] = None, heading: Optional[str] = None,
                     mediadirs: Optional[List[str]] = None, mainbody: Optional[Union[str, Callable[[], str]]] = None,
                     exclude_in_toc: Optional[List[str]] = None, include_in_toc: Optional[List[str]] = None):
            """
            :param format: The output format (one of :py:class:`ExportFormat`).
            :param name: The export definition name. Finding definitions by name is possible with `exports.getbyname`.
            :param heading: The title.
            :param mediadirs: Directories for additional media files like images used in the documentation.
            :param mainbody: Additional text to append to the selected documentation.
            :param exclude_in_toc: List of piece names to exclude for the table of contents. Default: Exclude none.
            :param include_in_toc: List of piece names to include for the table of contents. Default: Include all.
            """
            self.format = format
            self.name = name
            self.heading = heading
            self.mediadirs = mediadirs or []
            self.mainbody = mainbody
            self.exclude_in_toc = exclude_in_toc
            self.include_in_toc = include_in_toc

        def generate(self, targetpath: str) -> None:
            """
            Generates the content.

            :param targetpath: The content destination path.
            """
            Internals.generate_raw(head1=universe.name, targetpath=targetpath, outputformat=self.format,
                                   mainbody=(self.mainbody() if callable(self.mainbody) else self.mainbody) or "",
                                   short_desc=self.heading, mediadirs=self.mediadirs,
                                   exclude_in_toc=self.exclude_in_toc, include_in_toc=self.include_in_toc)

    class AbstractFileExport(Export):
        """
        Abstract base class for exports, which write documentation content to files.
        """

        # noinspection PyShadowingBuiltins
        def __init__(self, outputfile: str, **kwargs):
            """
            :param outputfile: The content destination path.
            """
            super().__init__(**kwargs)
            self.outputfile = outputfile

    @staticmethod
    def generate_homepage(*, head1: str, short_desc: str, targetdir: str, sections: List[Tuple[str, str, str]],
                          basecolor: Optional[Tuple[float, float, float]] = None,
                          aliases: Optional[Dict[str, str]] = None) -> str:
        """
        Writes the project homepage. You find it afterwards in :file:`index.html` (and tons of other files).

        :param head1: The first heading (typically the project name).
        :param short_desc: The additional heading description (typically a one-liner).
        :param targetdir: The directory for the html output.
        :param basecolor: The basecolor for colouring the output.
        :param sections: A list of content tuples.
        :param aliases: Custom command aliases.
        """
        page = universe.docgen.heading("Welcome!")
        for sectionkey, sectiontext, sectionhead in sections:
            page += (f"{universe.docgen.heading(sectionhead, variant='^', sub=True, anchor=sectionkey)}\n"
                     f"{sectiontext}\n\n")
        Internals.generate_raw(head1=head1, targetpath=targetdir, outputformat=ExportFormat.HTML,
                               mainbody=page, basecolor=basecolor, homepagemode=True, short_desc=short_desc,
                               aliases=aliases, include_in_toc=[])
        return targetdir

    @staticmethod
    def generate_raw(*, targetpath: str, outputformat: str, mainbody: Optional[str] = None, head1: Optional[str] = None,
                     basecolor: Optional[Tuple[float, float, float]] = None,
                     documentations: Optional[List[Tuple[str, str]]] = None, short_desc: Optional[str] = None,
                     homepagemode: bool = False, mediadirs: Optional[List[str]] = None,
                     aliases: Optional[Dict[str, str]] = None, exclude_in_toc: Optional[List[str]] = None,
                     include_in_toc: Optional[List[str]] = None) -> None:
        """
        Generates some documentation or output. Mostly used by some higher-level functions in this module.

        :param targetpath: The directory for the output.
        :param outputformat: The desired output format. See :py:class:`ExportFormat`.
        :param mainbody: Auxiliary rst source code for additional content.
        :param head1: The first heading.
        :param basecolor: The basecolor for colouring the output.
        :param documentations: An Anise documentations tuple for further pages that do not exist in the sourcecode
                               (e.g. the readme sources) as added by :samp:`anise.features.documentation.pool.add`.
        :param short_desc: The additional heading description (typically a one-liner) (html output only).
        :param homepagemode: If to generate html output for a homepage; styling differs a bit, sidebar behavior, ...
                             (html output only).
        :param mediadirs: Directories for additional media files like images used in the documentation.
        :param aliases: Custom command aliases.
        :param exclude_in_toc: List of piece names to exclude for the table of contents. Default: Exclude none.
        :param include_in_toc: List of piece names to include for the table of contents. Default: Include all.
        """
        pieces = list(universe.documentation.pieces.get_pieces())
        if head1 is None:
            head1 = universe.name
        basecolor = basecolor or universe.base.basecolor
        if documentations is None:
            documentations = universe.documentation.pool.gettexts()
        with anise.framework.files.TempDir() as tempd:
            for dk, dt in documentations:
                fdk = f'{tempd.path}/{dk}.rst'
                with open(fdk, 'w') as f:
                    f.write(dt)
                pieces.append(universe.docrender.RstTextsPiece(dk, fdk))
            head = f"{head1} - {short_desc}" if short_desc else head1
            if outputformat == ExportFormat.HTML:
                head = head1
                outputspec = universe.docrender.HtmlOutputSpec(targetpath, homepagemode=homepagemode,
                                                               title=head1, short_desc=short_desc,
                                                               theme=universe.documentation.htmltheme,
                                                               masterlink=universe.base.websiteurl)
            elif outputformat == ExportFormat.PDF:
                outputspec = universe.docrender.PdfOutputSpec(targetpath)
            elif outputformat == ExportFormat.Plaintext:
                outputspec = universe.docrender.PlaintextOutputSpec(targetpath)
            else:
                raise anise.framework.exceptions.BadInputError(f"Bad outputformat: {outputformat}")
            author = ", ".join(universe.base.authors)
            pieces.sort(key=lambda x: x.sortidx)
            universe.docrender.render(pieces, outputspec, head, author=author, mainbody=mainbody, basecolor=basecolor,
                                      mediadirs=mediadirs, shortsnippets=aliases,
                                      include_in_toc=include_in_toc, exclude_in_toc=exclude_in_toc)


class ExportFormat:
    """
    Enumeration of output formats.
    """

    #: Plaintext documentation export.
    Plaintext = "Plaintext"

    #: PDF documentation export.
    PDF = "PDF"

    #: HTML documentation export.
    HTML = "HTML"


class HomepageExport(Internals.Export):
    """
    Describes an export of an Anise documentation directly into the homepage.
    """

    def __init__(self, key: str, index: int = 100000, **kwargs):
        """
        :param key: One of the keys stored in pool. Defines which documentation to export.
        :param index: A position number. Used for ordering.
        """
        super().__init__(ExportFormat.HTML, **kwargs)
        self.key = key
        self.index = index


class PackageExport(Internals.AbstractFileExport):
    """
    Describes an export of an Anise documentation into the distribution package(s).
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SourceExport(Internals.AbstractFileExport):
    """
    Describes an export of an Anise documentation back into the source tree.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
