# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for handling the project's license.
"""

from typing import *
import datetime
import inspect
import os

import anise.framework.features
import anise.utils

from anise.framework.projects import universe

packages = anise.framework.features.loadfeature("packages").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(packages.HOOK_ENRICHRAWPACKAGE)
    def _writelicense(destdir):
        _license = universe.base.license
        if _license is not None:
            with universe.executionscope.newsubscope("Adding license file"):
                anise.utils.basic.writetofile(destdir + "/LICENSE", _license.getlicensetext())

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION, provides=["PRE"])
    def _instantiatelicense():
        if inspect.isclass(universe.base.license):
            universe.base.license = universe.base.license()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_homepage_section():

        def my_homepage_section(**_):
            return f"""{universe.name} is distributed under the terms of the `{universe.base.license.name} <license>`_ 
license. This also affects all included files without a license header (non-source files like images), unless
they are explicitly mentioned as third-party content.
Read the :ref:`Dependencies<dldependencies>` section for included third-party stuff.
    """

        universe.homepage.sections.add(2000, "license", "License", my_homepage_section)


class BaseLicense:
    """
    Base class for licenses. This interface is probably used by many features.
    """

    def __init__(self, namedebian: Optional[str] = None, namepythonsetuputils: Optional[str] = None,
                 name: Optional[str] = None, filename: Optional[str] = None):
        """
        :param namedebian: The name of the license in Debian packages.
        :param namepythonsetuputils: The name of the license on Python packages.
        :param name: The name of the license.
        :param filename: The path to the file with the license text.
        """
        self.name = name if (name is not None) else self.__class__.__name__
        self.namedebian = namedebian or self.name
        self.namepythonsetuputils = namepythonsetuputils or self.name
        self.filename = filename

    def getfilename(self) -> str:
        """
        Returns the full path of the license text file.
        """
        if self.filename is None:
            self.filename = f"//{type(self).__name__}"
        if self.filename.startswith("//"):
            p = f"{universe.specialpaths.anisedatadir}/licenses/{self.filename}"
        elif self.filename.startswith("/"):
            p = self.filename
        else:
            p = f"{universe.specialpaths.projectdir}/{self.filename}"
        return os.path.abspath(p)

    def getlicensetext(self) -> str:
        """
        Returns the full license text.
        """
        with open(self.getfilename(), "r") as f:
            licensetext = f.read()
        authors = " and ".join(([", ".join(universe.base.authors[0:-1])] if len(universe.base.authors) > 1 else [])
                               + [universe.base.authors[-1]])
        return licensetext \
            .replace("{{{year}}}", str(datetime.datetime.now().year)) \
            .replace("{{{authors}}}", authors) \
            .replace("{{{name}}}", universe.name) \
            .replace("{{{contributors}}}", "")


class AGPLv3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="AGPL-3", namepythonsetuputils="AGPL-3")


class GPLv3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="GPL-3", namepythonsetuputils="GPL-3")


class GPLv3OrHigher(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="GPL-3+", namepythonsetuputils="GPL-3+")
        self.name = GPLv3().name + " or higher"

    def getfilename(self):
        return GPLv3().getfilename()


class AFLv3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="AFL-3", namepythonsetuputils="AFL-3")


class Apache2(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="Apache-2", namepythonsetuputils="Apache-2")


class BSD3clause(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="BSD-3-clause", namepythonsetuputils="BSD-3-clause", name="BSD-3-clause")


class BSD2clause(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="BSD-2-clause", namepythonsetuputils="BSD-2-clause", name="BSD-2-clause")


class LGPLv3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="LGPL-3", namepythonsetuputils="LGPL-3")


class MPLv11(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="MPL-1.1", namepythonsetuputils="MPL-1.1", name="MPL-1.1")


class MPLv2(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="MPL-2", namepythonsetuputils="MPL-2", name="MPL-2.0")


class CcBy3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC-BY-3", namepythonsetuputils="CC-BY-3", name="CC-BY-3")


class CcBySa3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC-BY-SA-3", namepythonsetuputils="CC-BY-SA-3", name="CC-BY-SA-3")


class CcByNd3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC-BY-ND-3", namepythonsetuputils="CC-BY-ND-3", name="CC-BY-ND-3")


class CcByNc3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC-BY-NC-3", namepythonsetuputils="CC-BY-NC-3", name="CC-BY-NC-3")


class CcByNcSa3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC-BY-NC-SA-3", namepythonsetuputils="CC-BY-NC-SA-3", name="CC-BY-NC-SA-3")


class CcByNcNd3(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC-BY-NC-ND-3", namepythonsetuputils="CC-BY-NC-ND-3", name="CC-BY-NC-ND-3")


class Cc0v1(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="CC0-1", namepythonsetuputils="CC0-1", name="CC0-1")


class Artistic1(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="Artistic-1", namepythonsetuputils="Artistic-1")


class MIT(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="MIT", namepythonsetuputils="MIT")


class PublicDomain(BaseLicense):
    def __init__(self):
        super().__init__(namedebian="public-domain", namepythonsetuputils="public-domain")
