# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for specifying, generating and processing image galleries. It is typically used as part of the project
homepage.
"""

from typing import *
import os
import shutil

import anise.framework.features

from anise.framework.projects import universe


class Internals:

    class Pool:
        """
        Storage of all image galleries.
        """

        def __init__(self):
            self.list = []

        def add(self, gallery: 'MediaGallery', name: str = "") -> None:
            """
            Adds an image gallery.

            :param gallery: A media gallery specification.
            :param name: Gallery name.
            """
            self.list.append((name, gallery))

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.mediagalleries.pool = Internals.Pool()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _add_homepage_section():

        def my_homepage_section(homepage, **_):
            if len(universe.mediagalleries.pool.list) > 0:
                galleries = ""
                i = 0
                for name, gallery in universe.mediagalleries.pool.list:
                    galleries += universe.docgen.heading(name, variant="~")
                    galleries += ".. rst-class:: anisedoc-mediagallery\n\n"
                    ngallerydir = f"gallery{i}/"
                    cgallery = gallery.cloneto(homepage.path + "/" + ngallerydir)
                    for img in cgallery.getmedialist():
                        mtitle = img.description.replace("\n", " ").replace('"', "''")
                        mpath = ngallerydir + img.path
                        galleries += f"`{mtitle} <{mpath}#{img.mediatype}>`__\n"
                    i += 1
                return galleries

        universe.homepage.sections.add(6500, "gallery", "Gallery", my_homepage_section)


class Media:
    """
    One image in a gallery.
    """

    def __init__(self, path: str, description: str, mediatype: str):
        """
        :param path: The image file path.
        :param description: Description text.
        :param mediatype: The media type. Something like :samp:`'video'` or :samp:`'image'`.
        """
        self.path = path
        self.description = description
        self.mediatype = mediatype


class MediaGallery:
    """
    An image gallery.
    """

    def __init__(self, path: Optional[str] = None, paths: Optional[str] = None):
        """
        :param path: The directory path where the image gallery is stored.
        :param paths: The directory paths where the image gallery is stored.
        """
        paths = paths or []
        if path is not None:
            paths.append(path)
        self.paths = []
        for path in paths:
            while path.endswith("/"):
                path = path[:-1]
            self.paths.append(path)

    def _matchfile(self, name: str) -> Optional[str]:
        for pat in [".png", ".jpg", ".jpeg"]:
            if name.endswith(pat):
                return "image"
        for pat in [".mp4", ".webm"]:
            if name.endswith(pat):
                return "video"

    def getmedialist(self) -> List['Media']:
        """
        Returns the list of media in the gallery.
        """
        result = []
        for path in self.paths:
            isabsolute = path.startswith("/")
            if isabsolute:
                fullpath = path
            else:
                fullpath = universe.specialpaths.projectdir + "/" + path
            for img in os.listdir(fullpath):
                mediatype = self._matchfile(img)
                if mediatype:
                    fimgdesc = path + "/" + img + ".txt"
                    desc = ""
                    if os.path.isfile(fimgdesc):
                        with open(fimgdesc, "r") as f:
                            desc = f.read()
                    result.append(Media(img, desc, mediatype))
        result.sort(key=lambda x: x.path)
        return result

    def cloneto(self, destpath: str) -> 'MediaGallery':
        """
        Clones the media gallery directory to a different place.

        :param destpath: The destination where to copy the image gallery data. Absolute path.
        :return: The new MediaGallery instance pointing to the cloned directory.
        """
        os.makedirs(destpath, exist_ok=True)
        for path in self.paths:
            isabsolute = path.startswith("/")
            if isabsolute:
                fullpath = path
            else:
                fullpath = universe.specialpaths.projectdir + "/" + path
            for img in os.listdir(fullpath):
                if self._matchfile(img) or img.endswith(".txt"):
                    shutil.copyfile(path + "/" + img, destpath + "/" + img)
        return MediaGallery(destpath)
