# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Feature for creation of Flatpak images.
"""

from typing import *
import os
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

milieus = anise.framework.features.loadfeature("milieus").featuremodule
packagesdebian = anise.framework.features.loadfeature("packages.debian").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.packages.flatpak.repository = getattr(universe.packages.flatpak, "repository", None)
        universe.packages.flatpak.gpg_signkeyid = getattr(universe.packages.flatpak, "gpg_signkeyid", "")
        universe.packages.flatpak.gpg_homedir = getattr(universe.packages.flatpak, "gpg_homedir", "")
        universe.packages.flatpak.gpg_exportto = getattr(universe.packages.flatpak, "gpg_exportto", "")


class MenuEntry(packagesdebian.MenuEntry):
    """
    Specification for one menu entry added by a Flatpak.
    """

    def __init__(self, command="", *a, **b):
        """
        See more about parameters in the superclass.
        """
        super().__init__(command=command, *a, **b)


class Metafile(anise.framework.files.Filestructure):
    """
    An abstract description for a Flatpak meta file for easier access (e.g. a .flatpakref).
    
    Use one of the non-abstract subclasses.
    """

    def __init__(self, filename: Optional[str]):
        super().__init__(None)
        self.filename = filename
        self.fullname = None
        self.repobaseurl = None
        self.repourl = None
        self.gpgkey = None

    def set(self, fullname: str, repobaseurl: str, repourl: str, gpgkey: bytes):
        """
        Called by the infrastructure for setting some data.
        |infrastructure|
        """
        self.fullname = fullname
        self.repobaseurl = repobaseurl
        self.repourl = repourl
        self.gpgkey = gpgkey

    def _initialize(self):
        self.filename = self.filename or f"{universe.name}.flatpakref"
        self._path = f"{anise.framework.files.TempDir().path}/{self.filename}"
        os.makedirs(os.path.dirname(self._path), exist_ok=True)
        with open(self._path, "w") as f:
            f.write(self.generate())

    def generate(self) -> str:
        """
        Generates and returns the metafile content.
        |dooverride|
        """
        raise NotImplementedError()


class FlatpakRef(Metafile):
    """
    Description for creation of a Flatpak application reference (.flatpakref) file.
    """

    def __init__(self, filename: Optional[str] = None, title: Optional[str] = None, branch: str = "master",
                 runtimerepourl: Optional[str] = None):
        super().__init__(filename)
        self.title = title
        self.branch = branch
        self.runtimerepourl = runtimerepourl

    def generate(self):
        title = self.title or universe.name
        res = {"Title": title, "Name": self.fullname, "Branch": self.branch, "Url": self.repourl,
               "IsRuntime": "False"}
        if self.gpgkey:
            res["GPGKey"] = self.gpgkey.decode()
        if self.runtimerepourl:
            res["RuntimeRepo"] = self.runtimerepourl
        self._data["description"] = (f"or install it with just:\n\n.. code-block:: sh\n\n" 
                                     f"  $ flatpak install --user --from {self.repobaseurl}/{self.filename}\n\n")
        return "[Flatpak Ref]\n" + "\n".join([f"{x}={res[x]}" for x in res])


class FlatpakMilieu(milieus.Milieu):
    """
    Build environment for a new Flatpak image.
    
    Typically this is used automatically within :py:func:`flatpakpackage`.
    
    See :py:class:`anise.features.milieus.MilieuComposite` for general infos.
    """

    def __init__(self, name: str, sdk: str = "org.freedesktop.Sdk", platform: str = "org.freedesktop.Platform",
                 kitversion: str = None, fullname: str = None,
                 sockets: Iterable[str] = ("x11",), filesystems: Iterable[str] = ("home",),
                 shares: Iterable[str] = ("network",),
                 command: Optional[str] = None,
                 menuentries: Optional[List[MenuEntry]] = None, environment: Optional[Dict[str, str]] = None,
                 flatpakpath: str = "flatpak", gpg_signkeyid: str = "", gpg_homedir: str = "", gpg_exportto: str = ""):
        """
        :param name: The Flatpak image short name.
        :param sdk: The sdk name, like :samp:`"org.kde.Sdk"`.
        :param platform: The platform name, like :samp:`"org.kde.Platform"`.
        :param kitversion: The version of the sdk and platform as string.
        :param fullname: The full image name (dotted string).
        :param sockets: The flatpak sockets to allow the application to access.
        :param filesystems: The flatpak filesystems to allow the application to access.
        :param shares: The flatpak shares to allow the application to access.
        :param command: The optional command string. Must be a file in :file:`/app/bin` inside the sandbox.
        :param menuentries: The list of menu entries this image should add.
        :param environment: Environment variables for within the sandbox.
        :param flatpakpath: The path to the flatpak tool.
        :param gpg_signkeyid: The gpg key id used for signing products.
        :param gpg_homedir: For gpg_sign, this optionally specifies the directory that contains that key.
        :param gpg_exportto: The optional target file name for making the gpg key available for download.
        """
        milieus.Milieu.__init__(self)
        self.name = name
        self.sdk = sdk
        self.platform = platform
        self.kitversion = kitversion
        self.fullname = fullname
        self.sockets = list(sockets)
        self.filesystems = list(filesystems)
        self.shares = list(shares)
        self.command = command
        self.menuentries = menuentries or []
        self.environment = environment or {}
        self.gpg_signkeyid = gpg_signkeyid
        self.gpg_homedir = gpg_homedir
        self.gpg_exportto = gpg_exportto
        self.pgpg = (["--gpg-sign=" + gpg_signkeyid] if gpg_signkeyid else []) + \
                    (["--gpg-homedir=" + gpg_homedir] if gpg_homedir else [])
        self.flatpakpath = flatpakpath
        self.flatpakcmd = ["{flatpakpath}", "[[args]]"]
        self._envrootpath = None

    def initialize(self):
        super().initialize()
        self._envrootpath = self.createtempdir(self.name).path
        p = ["build-init", self._envrootpath, self.fullname, self.sdk, self.platform]
        if self.kitversion:
            p.append(self.kitversion)
        self.call_flatpak(*p)

    def finish_build(self) -> None:
        p = ["build-finish", self._envrootpath]
        if self.command:
            p.append("--command=" + self.command)
        for s in self.sockets:
            p.append("--socket=" + s)
        for s in self.filesystems:
            p.append("--filesystem=" + s)
        for s in self.shares:
            p.append("--share=" + s)
        for ek in self.environment:
            p.append("--env=" + ek + "=" + self.environment[ek])
        self.call_flatpak(*p)
        sharedappspath = self._envrootpath + "/export/share/applications"
        sharedpixmapspath = self._envrootpath + "/export/share/icons"
        os.makedirs(sharedappspath)
        os.makedirs(sharedpixmapspath)
        for menuentry in self.menuentries:
            iconsrc = universe.specialpaths.projectdir + "/" + menuentry.icon
            iconfname = self.fullname + "." + menuentry.name + ".png"
            shutil.copyfile(iconsrc, sharedpixmapspath + "/" + iconfname)
            with open(sharedappspath + "/" + self.fullname + "." + menuentry.name + ".desktop", "w") as f:
                f.write(f"[Desktop Entry]\n"
                        f"Name={menuentry.title}\n"
                        f"Exec={menuentry.command}\n"
                        f"Terminal={'false' if menuentry.gui else 'true'}\n"
                        f"Type=Application\n"
                        f"Categories={menuentry.category[1]};\n"
                        f"Icon={iconfname}\n")

    def call(self, *args: str, inside_sandbox: bool = True, **kwargs):
        if not inside_sandbox:
            return super().call(*args, **kwargs)
        cmd = [x.format(**self.__dict__) for x in self.flatpakcmd]
        i = cmd.index("[[args]]")
        if getattr(kwargs, "shell", False):
            args = ["bash", "-c", *args]
            kwargs["shell"] = False
        cmd = cmd[:i] + ["build", "--share=network", self._envrootpath] + list(args) + cmd[i+1:]
        return super().call(*cmd, **kwargs)

    def createtempdir(self, dirname: Optional[str] = None) -> anise.framework.files.TempDir:
        return anise.framework.files.TempDir(dirname=dirname, temprootpath=os.path.expanduser("~/"),
                                             namepattern=".anise.tmp.flatpakbuild.{pid}.{i}")

    def call_flatpak(self, *args: str, raise_on_errors: Optional[str] = "unable to run flatpak",
                     **kwargs) -> Tuple[int, Any]:
        cmd = [x.format(**self.__dict__) for x in self.flatpakcmd]
        i = cmd.index("[[args]]")
        cmd = cmd[:i] + list(args) + cmd[i+1:]
        return self.call(*cmd, raise_on_errors=raise_on_errors, inside_sandbox=False, **kwargs)

    def deploy_to_repository(self, repository: str, src: Optional[str] = None) -> None:
        ffname = src or self._envrootpath
        self.call_flatpak(*(["build-export"] + self.pgpg + [repository, ffname]))
        self.call_flatpak(*(["build-update-repo"] + self.pgpg + [repository]))


def flatpakpackage(source: anise.framework.files.Filestructure, fullname: Optional[str] = None,
                   repository: Optional[str] = None, repository_external_baseurl: Optional[str] = None,
                   metafiles: Optional[List[Metafile]] = None,
                   gpg_signkeyid: str = "", gpg_homedir: str = "", gpg_exportto: str = "",
                   **milieuopts) -> anise.framework.files.Filestructure:
    """
    Builds a Flatpak and upload it to a repository.
    
    See also :py:class:`FlatpakMilieu` for more parameters.

    :param source: A source file structure.
    :param fullname: The full image name (dotted string).
    :param repository: The repository to upload the flatpak to instead of deploying it inside the homepage data.
    :param repository_external_baseurl: The external base url of the repository.
    :param metafiles: Optional list of meta files to create for easier access.
    :param gpg_signkeyid: The gpg key id used for signing products.
    :param gpg_homedir: For gpg_sign, this optionally specifies the directory that contains that key.
    :param gpg_exportto: The optional target file name for making the gpg key available for download.
    :param milieuopts: Additional options for the Flatpak milieu (see FlatpakMilieu).
    """
    with universe.executionscope.newsubscope("Creating flatpak image"):
        name = universe.name
        _oldbe = universe.milieus.defaultmilieu
        try:
            fullname = (fullname or universe.packages.flatpak.fullnamepattern).format(**locals())
            repository = repository or universe.packages.flatpak.repository
            metafiles = metafiles or []
            gpg_signkeyid = gpg_signkeyid or universe.packages.flatpak.gpg_signkeyid
            gpg_homedir = gpg_homedir or universe.packages.flatpak.gpg_homedir
            gpg_exportto = (gpg_exportto or universe.packages.flatpak.gpg_exportto).format(**locals())
            _pgpg = (["--gpg-sign=" + gpg_signkeyid] if gpg_signkeyid else []) + \
                    (["--gpg-homedir=" + gpg_homedir] if gpg_homedir else [])
            milieuopts.update({'class': FlatpakMilieu, 'name': name, 'fullname': fullname,
                               'gpg_signkeyid': gpg_signkeyid, 'gpg_homedir': gpg_homedir,
                               'gpg_exportto': gpg_exportto})
            if isinstance(universe.milieus.defaultmilieu, list):
                milieu = universe.milieus.get([milieuopts] + universe.milieus.defaultmilieu)
            else:
                milieu = universe.milieus.get([milieuopts, universe.milieus.defaultmilieu])
            universe.milieus.defaultmilieu = milieu
            if milieu.call("flatpak", "--version", inside_sandbox=False)[0] != 0:
                raise anise.framework.exceptions.RequirementsMissingError("'flatpak' seems to be unavailable")
            repository_external_baseurl = (repository_external_baseurl or
                                           universe.packages.flatpak.repository_external_baseurl).format(**locals())
            containerdir = milieu.createtempdir().path
            # noinspection PyProtectedMember
            source.dl(milieu._envrootpath)
            milieu.finish_build()
            newreponame = None
            if repository is None:
                repository = newreponame = milieu.createtempdir(f"{name}_anise.{os.getpid()}-repo").path
            repository_external_url = f"{repository_external_baseurl}/flatpak"
            milieu.deploy_to_repository(repository)
            res = anise.framework.files.TempDir()
            if newreponame:
                anise.framework.files.Filestructure(newreponame).dl(f"{res.path}/flatpak")
            res.setdata("invisible_on_homepage", True)
            res.setdata("packagestore_only_latest", True)
            gpgkey = universe.gpg.get_key(gpg_signkeyid, homedir=gpg_homedir) if gpg_signkeyid else None
            for metafile in metafiles:
                metafile.set(fullname, repository_external_baseurl, repository_external_url, gpgkey)
                metafile.setdata("packagestore_only_latest", True)
                universe.distributables.currentgroup.sources.append(metafile)
            if gpg_exportto and gpg_signkeyid:
                rgpgkey = universe.gpg.get_key(gpg_signkeyid, homedir=gpg_homedir, withbase64=False)
                gpgff = f"{anise.framework.files.TempDir().path}/{gpg_exportto}"
                with open(gpgff, "wb") as f:
                    f.write(rgpgkey)
                gpgfile = anise.framework.files.Filestructure(gpgff)
                gpgfile.setdata("packagestore_only_latest", True)
                gpgfile.setdata("description", "Use the .gpg for optional repository signature verification.")
                universe.distributables.currentgroup.sources.append(gpgfile)
            def _distrihtml(icon, text):
                return (f".. container:: anisedoc-infobutton\n\n"
                        f"  .. image:: aniseicons/{icon}.png\n\n"
                        f"  .. hint:: \n    {text}\n\n")
            res.setdata("description",
                        (f"Install {name} as Flatpak on most modern Linux systems by "
                         f"installing application ':samp:`{fullname}`' from the "
                         f"repository ':samp:`{repository_external_url}`'.\n\n" +
                         _distrihtml("archlinux", "Install `$ sudo pacman -S flatpak` "
                                                  "before you can use Flatpak in Arch Linux.") +
                         _distrihtml("centos", "Install `$ yum install flatpak` as root "
                                               "before you can use Flatpak in CentOS.") +
                         _distrihtml("debian", "Install `$ apt install flatpak` as root "
                                               "before you can use Flatpak in Debian.") +
                         _distrihtml("fedora", "Flatpak is directly usable in Fedora without further work.") +
                         _distrihtml("gentoo", "Install `https://github.com/fosero/flatpak-overlay` "
                                               "before you can use Flatpak in Gentoo.") +
                         _distrihtml("opensuse", "Install `$ sudo zypper install flatpak` "
                                                 "before you can use Flatpak in openSUSE.") +
                         _distrihtml("redhat", "Install `$ yum install flatpak` as root "
                                               "before you can use Flatpak in Red Hat.") +
                         _distrihtml("ubuntu", "Install `$ sudo apt install flatpak` "
                                               "before you can use Flatpak in Ubuntu.") +
                         f".. rst-class:: anisedoc-infobutton-stop\n\n"
                         f"See above how to prepare Flatpak and then run:\n\n"
                         f".. code-block:: sh\n\n"
                         f"  $ flatpak remote-add --user --no-gpg-verify {name} {repository_external_url}\n"
                         f"  $ flatpak install --user {name} {fullname}\n"
                         f"  $ flatpak run {fullname}\n\n"))
            return res
        finally:
            universe.milieus.defaultmilieu = _oldbe
