# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

import anise.framework.engine
import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.framework.projects
from anise import utils

from anise.framework.projects import universe

diagnostics = anise.framework.features.loadfeature("diagnostics").featuremodule
projectdescriptioneditor = anise.framework.features.loadfeature("projectdescriptioneditor").featuremodule


def formatcodefrom(code):
    return code.strip() + "\n"


def formatcodeto(code):
    return "\n" + code.strip() + "\n        "


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS,
                                      provides=["PRE", "task"], requires=["manual"])
class SimpleTaskAssistantApplication(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self, forfeature=""):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage Tasks", "Create new task implementations here or change existing ones.", forfeature)

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add new task", self._add, icon="userfeedback/add.png")]

    def getentries(self, ais):
        return [x for x in ais.entries() if x.islet and x.key == "base.pytask"]

    def getentrylabel(self, ais, entry):
        return entry.getarg("taskname").value

    def _add(self, ais):
        tname = ""
        while tname == "":
            tname = SimpleTaskAssistantApplication.getnewkey(None, ais)
        if tname:
            sample = """# The current content is just an example. Feel free to remove it
# and implement your task here.

with universe.executionscope.newsubscope("Doing some crazy "+universe.name+" stuff"):
    answer = universe.ui.userfeedback.message("Hello World!", ["Foo", "Bar"])
    utils.logging.loginfo(answer)
            """
            newlet = anise.framework.projects.IntermediateStructure.Let(
                "base.pytask",
                code=anise.framework.projects.IntermediateStructure.Value.fromstring(formatcodeto(sample)),
                **{"taskname": tname})
            ais.addentry(newlet)

    @staticmethod
    def getnewkey(defaultname, ais):
        if defaultname is None:
            msg = "Please enter the name for the new task."
            defaultname = "mytask"
            i = 1
            while [x for x in ais.entries()
                   if isinstance(x, anise.framework.projects.IntermediateStructure.Let)
                   and x.key == "base.pytask" and x.getarg("taskname").value == defaultname]:
                defaultname = "mytask" + str(i)
                i += 1
        else:
            msg = "Please enter the new task name."
        r = universe.ui.userfeedback.input(msg=msg, default=defaultname)
        if r is None:
            return None
        else:
            return anise.framework.projects.IntermediateStructure.Value.fromstring(r)


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS,
                                      provides=["PRE"], requires=["task", "manual"])
class SimpleScriptAssistantApplication(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    sample_function = """def on_{ttname}({signature}):
    # This is sample content. Change it as needed!
{body}
"""

    sample_class = """class HookHandler_{ttname}{sbaseclass}:
    # This is sample content. Change it as needed!

{body}
"""

    def __init__(self, forfeature=""):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage Hook Scripts",
            "Create new hook handler implementations here or change existing ones. This is more advanced and requires "
            "to study the manual before.", forfeature)

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add new hook script", self._add, icon="userfeedback/add.png")]

    def getentries(self, ais):
        return [x for x in ais.entries() if x.islet and x.key == "base.pyscript"]

    def getentrylabel(self, ais, entry):
        i = [x for x in ais.entries()
             if isinstance(x, anise.framework.projects.IntermediateStructure.Let)
             and x.key == "base.pyscript"].index(entry) + 1
        return "#{} ({})".format(i, entry.getarg("hook").oref)

    def _add(self, ais):
        tnod, sampletype = SimpleScriptAssistantApplication.getnewkey(None)
        if tnod:
            sample = self.getsample(tnod, sampletype)
            if sample is not None:
                newlet = anise.framework.projects.IntermediateStructure.Let(
                    "base.pyscript",
                    code=anise.framework.projects.IntermediateStructure.Value.fromstring(formatcodeto(sample)),
                    **{"hook": tnod})
                ais.addentry(newlet)

    @staticmethod
    def _getallhooks():
        hooks = [(None, "Specify hook reference manually", None, None),
                 ("engine", "HOOK_AFTER_PREPARATION", "engine.HOOK_AFTER_PREPARATION",
                  anise.framework.engine.HOOK_AFTER_PREPARATION),
                 ("engine", "HOOK_SHUTDOWN", "engine.HOOK_SHUTDOWN", anise.framework.engine.HOOK_SHUTDOWN)]
        for feat in anise.framework.features.getfeatures():
            for ohook in feat.hooks:
                hooks.append((feat.name, ohook, feat.name + "." + ohook, feat.hooks[ohook]))
        tmphooks = [((-100 if x[0] is None else x[3].interesting), x) for x in hooks]
        tmphooks.sort()
        return [x[1] for x in tmphooks]

    @staticmethod
    def getnewkey(defaultname):
        if defaultname is None:
            msg = "Please specify the hook you want to a handler for."
            defaultname = "engine.HOOK_AFTER_PREPARATION"
        else:
            msg = "Please specify the new hook."
        sampletype = None
        hooks = [x for x in SimpleScriptAssistantApplication._getallhooks() if (not x[3]) or (not x[3].hide)]
        rc = universe.ui.userfeedback.choice(msg=msg, choices=[(x[1]+" in "+x[0]) if x[0] else x[1] for x in hooks])
        r = None
        if rc is not None:
            featname, hookname, hookfullname, hookobj = hooks[rc]
            if hookobj is None:
                r = universe.ui.userfeedback.input(msg="Please manually specify the hook reference.",
                                                   default=defaultname)
            else:
                if hookobj.hookkind == anise.framework.features.HookKind.CLASS:
                    sampletype = SimpleScriptAssistantApplication.sample_class
                elif hookobj.hookkind == anise.framework.features.HookKind.FUNCTION:
                    sampletype = SimpleScriptAssistantApplication.sample_function
                else:
                    raise anise.framework.exceptions.ImplementationError("Invalid hook kind: " + repr(hookobj.hookkind))
                r = hookfullname
        if r is None:
            return None, None
        else:
            return anise.framework.projects.IntermediateStructure.Value("objectref", oref=r), sampletype

    def getsample(self, tname, sampletype):
        if sampletype is None:
            r = universe.ui.userfeedback.message("Which type of hook handler do you want to create?",
                                                 ["Class-type handler", "Function-type handler"])
            if r == 0:
                sampletype = SimpleScriptAssistantApplication.sample_class
            elif r == 1:
                sampletype = SimpleScriptAssistantApplication.sample_function
        ttname = tname.oref
        ildot = ttname.rfind(".")
        if ildot > -1:
            ttname = ttname[ildot+1:]
        if ttname.startswith("HOOK_"):
            ttname = ttname[5:]
        hookobj = [x for x in self._getallhooks() if x[2] == tname.oref][0][3]
        res = sampletype
        if res == SimpleScriptAssistantApplication.sample_function:
            fbody = []
            for prm in hookobj.wizard_functionparams:
                prm = prm.strip()
                while prm.startswith("*"):
                    prm = prm[1:]
                if "=" in prm:
                    prm = prm[:prm.find("=")].strip()
                fbody.append("utils.logging.loginfo('{prm}: ' + str({prm}))\n".format(prm=prm))
            for prm in hookobj.wizard_functionadditionallog:
                fbody.append("utils.logging.loginfo('{prm}: ' + str({prm}))\n".format(prm=prm))
            if hookobj.wizard_functionbody:
                fbody += hookobj.wizard_functionbody.replace("\r", "").split("\n")
            if not fbody:
                fbody.append("utils.logging.loginfo('Hello World.')")
            res = res.format(ttname=ttname, signature=", ".join([x.strip() for x in hookobj.wizard_functionparams]),
                             body="\n".join(["    "+x for x in fbody]))
        elif res == SimpleScriptAssistantApplication.sample_class:
            if hookobj.wizard_classbody:
                fbody = hookobj.wizard_classbody.strip().replace("\r", "").split("\n")
            else:
                fbody = ["def __init__(self):", "    pass"]
            sbaseclass = "({})".format(hookobj.wizard_classbase) if hookobj.wizard_classbase else ""
            res = res.format(ttname=ttname, sbaseclass=sbaseclass, body="\n".join(["    "+x for x in fbody]))
        else:
            raise anise.framework.exceptions.ImplementationError("Unknown sampletype.")
        if hookobj.__doc__:
            res += "\n\n" + "\n".join(["# "+x.strip() for x in hookobj.__doc__.split("\n")])
        for fnf in hookobj.wizard_loadfeatures:
            locf = "_" + fnf.replace(".", "")
            res = "{} = features.loadfeature('{}').featuremodule\n".format(locf, fnf) + res
        return res


# noinspection PyPep8Naming
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS, provides=["btask"])
class SimpleTaskAssistantApplication_InBase(SimpleTaskAssistantApplication):

    def __init__(self):
        super().__init__(forfeature="base")


# noinspection PyPep8Naming
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS, requires=["btask"])
class SimpleScriptAssistantApplication_InBase(SimpleScriptAssistantApplication):

    def __init__(self):
        super().__init__(forfeature="base")


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class EditCodeInlineCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Edit source code inline", icon="userfeedback/text.png")

    def visible(self, ais, node):
        return node.islet and node.key in ["base.pytask", "base.pyscript"]

    def execute(self, ais, node):
        sourcecode = formatcodefrom(node.getarg("code").value)
        newsourcecode = universe.ui.userfeedback.multilineinput(msg="Please modify the source code here and proceed.",
                                                                default=sourcecode)
        if newsourcecode is not None:
            node.getarg("code").value = formatcodeto(newsourcecode)


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class EditCodeExternalEditorCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Edit source code in external text editor", icon="userfeedback/text.png")

    def visible(self, ais, node):
        return node.islet and node.key in ["base.pytask", "base.pyscript"]

    def execute(self, ais, node):
        if universe.ui.userfeedback.message("In the following, an external program will allow you to edit the source "
                                            "code.\n\nIn order to apply your changes, you must also follow the "
                                            "instructions arising here afterwards!", image="userfeedback/warning.png",
                                            btns=["I understood, proceed", "Cancel"]) == 0:
            sourcecode = formatcodefrom(node.getarg("code").value)
            with anise.framework.files.TempDir() as tempdir:
                tmpfile = tempdir.path + "/anise.src.py"
                utils.basic.writetofile(tmpfile, sourcecode)
                utils.basic.openfile(tmpfile)
                universe.ui.userfeedback.message("An external program should open in some seconds. Please edit the "
                                                 "source code there and save it. Afterwards, please proceed here. Do "
                                                 "not proceed before you saved your changes!",
                                                 image="userfeedback/warning.png", btns=["Proceed"])
                nv = utils.basic.readfromfile(tmpfile)
            node.getarg("code").value = formatcodeto(nv)


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class RenameTaskCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Rename task")

    def visible(self, ais, node):
        return node.islet and node.key == "base.pytask"

    def execute(self, ais, node):
        oldname = node.getarg("taskname").value
        tname = ""
        while tname == "":
            tname = SimpleTaskAssistantApplication.getnewkey(oldname, ais)
        if tname:
            node.getarg("taskname").value = tname


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(
    projectdescriptioneditor.HOOK_GET_PROJECTDESCRIPTIONNODE_CUSTOMACTION_IMPLEMENTATIONS)
class ChangeHookCustomAction(projectdescriptioneditor.CustomAction):

    def __init__(self):
        super().__init__("Change hook")

    def visible(self, ais, node):
        return node.islet and node.key == "base.pyscript"

    def execute(self, ais, node):
        oldname = node.getarg("hook").oref
        tname = ""
        while tname == "":
            tname, sampletype = SimpleScriptAssistantApplication.getnewkey(oldname)
        if tname:
            node.getarg("hook").oref = tname
