# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

import anise.framework.exceptions
import anise.framework.features
import anise.framework.projects

from anise.framework.projects import universe

projectdescriptioneditor = anise.framework.features.loadfeature("projectdescriptioneditor").featuremodule


HOOK_GET_DATAINJECTION_GENERATORS = anise.framework.features.Hook(
    hide=True, doc="This hook is called for collecting datainjection generator implementations.")


class AbstractDataInjectionGenerator:

    def __init__(self):
        self.label = None

    def getfctref(self):
        raise anise.framework.exceptions.NotImplementedError()


@anise.framework.features.hookhandler(HOOK_GET_DATAINJECTION_GENERATORS)
class DataInjectionGeneratorCustom(AbstractDataInjectionGenerator):

    def __init__(self):
        AbstractDataInjectionGenerator.__init__(self)
        self.label = "Custom"

    def getfctref(self):
        return universe.ui.userfeedback.input("Please specify an objectref string pointing to a "
                                              "content generator function.")


@anise.framework.features.hookhandler(HOOK_GET_DATAINJECTION_GENERATORS)
class DataInjectionGeneratorPython(AbstractDataInjectionGenerator):

    def __init__(self):
        AbstractDataInjectionGenerator.__init__(self)
        self.label = "Python"

    def getfctref(self):
        return "datainjections.inject_python"


@anise.framework.features.hookhandler(HOOK_GET_DATAINJECTION_GENERATORS)
class DataInjectionGeneratorC(AbstractDataInjectionGenerator):

    def __init__(self):
        AbstractDataInjectionGenerator.__init__(self)
        self.label = "C / C++"

    def getfctref(self):
        return "datainjections.inject_c"


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS)
class ProjectDescriptionEditorAssistant(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage data injections", "Specify rules for generating additional files with dynamic content, like "
                                            "your project revision number or build time. Depending on how they are "
                                            "configured, they are generated at release time or when packages are "
                                            "created. Those files can be loaded at runtime of your software in order "
                                            "to retrieve those infos.", "datainjections")

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add data injection", self._add_datainjection,
                                                      icon="userfeedback/add.png")]

    def _add_datainjection(self, ais):
        doinsource = universe.ui.userfeedback.messageyesno("Should the injection take place directly in the "
                                                           "project directory?")
        outfile = universe.ui.userfeedback.input("Please specify the output file path.")
        if outfile is None:
            return
        digens = [x() for x in universe.gethooks(HOOK_GET_DATAINJECTION_GENERATORS)]
        digenchoices = [x.label for x in digens]
        digen = universe.ui.userfeedback.choice("Please choose a data injection.", digenchoices)
        if digen is None:
            return
        oref = digens[digen].getfctref()
        if oref is None:
            return
        newlet = anise.framework.projects.IntermediateStructure.Let(
            "datainjections.pool.add", outfile=anise.framework.projects.IntermediateStructure.Value.fromstring(outfile),
            do_in_source=anise.framework.projects.IntermediateStructure.Value("bool", v=doinsource),
            method=anise.framework.projects.IntermediateStructure.Value("objectref", oref=oref))
        ais.addentry(newlet)

    def getentries(self, ais):
        r = []
        for e in ais.entries():
            if isinstance(e, anise.framework.projects.IntermediateStructure.Let) \
                    and e.key == "datainjections.pool.add":
                r.append(e)
        return r

    def getentrylabel(self, ais, entry):
        r = "unknown"
        try:
            r = entry.getarg("outfile").value
            r += " / " + entry.getarg("method").oref
        except NameError:
            pass
        except KeyError:
            pass
        return r
