# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

import anise.framework.features
import anise.framework.projects

from anise.framework.projects import universe

projectdescriptioneditor = anise.framework.features.loadfeature("projectdescriptioneditor").featuremodule


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS)
class ProjectDescriptionEditorAssistantPool(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage documentation pool", "Specify how to generate documentation from which sources. Those "
                                               "documentations can be used by exporting it to some places.",
            "documentation")

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add documentation to pool", self._add_pool,
                                                      icon="userfeedback/add.png")]

    def _add_pool(self, ais):
        key = universe.ui.userfeedback.input("Please specify a documentation key you want to export to.", "MYDOCKEY")
        if key is None:
            return
        oref = universe.ui.userfeedback.input(
            "Please specify an objectref string pointing to the content generator function.",
            "lambda: 'A Horse is a horse, of course, of course.'")
        if oref is None:
            return
        vsrc = anise.framework.projects.IntermediateStructure.Value("objectref", oref=oref)
        newlet = anise.framework.projects.IntermediateStructure.Let(
            "documentation.pool.add", src=vsrc,
            key=anise.framework.projects.IntermediateStructure.Value.fromstring(key))
        ais.addentry(newlet)

    def getentries(self, ais):
        r = []
        for e in ais.entries():
            if isinstance(e, anise.framework.projects.IntermediateStructure.Let) \
                    and e.key == "documentation.pool.add":
                r.append(e)
        return r

    def getentrylabel(self, ais, entry):
        return entry.getarg("key").value


# noinspection PyUnusedLocal
@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS)
class ProjectDescriptionEditorAssistantExports(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage documentation exports", "Specify how to export documentation from the pool to which "
                                                  "destinations. The exports are typically executed for package "
                                                  "creation and for generation of the project homepage.",
            "documentation")

    def getcustomactions(self, ais):
        return [projectdescriptioneditor.CustomAction("Add documentation export", self._add_export,
                                                      icon="userfeedback/add.png")]

    def _add_export(self, ais):
        exportargs = {}
        exptypes = ["Source tree export", "Destination package export", "Homepage export"]
        exptype = universe.ui.userfeedback.choice("Please choose an license.", exptypes)
        if exptype == 0:
            exportargs["t"] = "documentation.SourceExport"
        elif exptype == 1:
            exportargs["t"] = "documentation.PackageExport"
        elif exptype == 2:
            exportargs["t"] = "documentation.HomepageExport"
        else:
            return
        if exptype in [0, 1]:
            outputfile = universe.ui.userfeedback.input("Please specify the destination path.", "/manual")
            if outputfile is None:
                return
            exportargs["outputfile"] = anise.framework.projects.IntermediateStructure.Value.fromstring(outputfile)
            expformats = ["Plaintext", "PDF", "HTML"]
            expformat = universe.ui.userfeedback.choice("Please choose an output format.", expformats)
            if expformat is None:
                return
            exportargs["format"] = anise.framework.projects.IntermediateStructure.Value.fromstring(
                expformats[expformat])
        key = universe.ui.userfeedback.input("Please specify a documentation key to export.", "FULLREADME")
        if key is None:
            return
        exportargs["key"] = anise.framework.projects.IntermediateStructure.Value.fromstring(key)
        heading = universe.ui.userfeedback.input("Please specify a heading", "Manual")
        if heading is None:
            return
        exportargs["heading"] = anise.framework.projects.IntermediateStructure.Value.fromstring(heading)
        newlet = anise.framework.projects.IntermediateStructure.Let("documentation.exports.add")
        newlet.setarg("export", anise.framework.projects.IntermediateStructure.Value(**exportargs))
        ais.addentry(newlet)

    def getentries(self, ais):
        r = []
        for e in ais.entries():
            if isinstance(e, anise.framework.projects.IntermediateStructure.Let) \
                    and e.key == "documentation.exports.add":
                r.append(e)
        return r

    def getentrylabel(self, ais, entry):
        return entry.getarg("export").getarg("key").value


@anise.framework.features.hookhandler(projectdescriptioneditor.HOOK_GET_FEATURE_ACTIONS)
class ProjectDescriptionEditorAssistantDevdoc(projectdescriptioneditor.ProjectDescriptionEditorAssistant):

    def __init__(self):
        projectdescriptioneditor.ProjectDescriptionEditorAssistant.__init__(
            self, "Manage developer documentation creation", "Specify if the README shall contain a developer "
                                                             "reference documentation build from the project source "
                                                             "code.", "documentation")

    def getcustomactions(self, ais):
        devdocisenabled = False
        for x in ais.entries():
            if isinstance(x, anise.framework.projects.IntermediateStructure.Let) and \
                            x.key == "documentation.has_developerdocumentation":
                devdocisenabled = True
        return [projectdescriptioneditor.CustomAction("Disable developer documentation", self._disable_devdoc) if
                devdocisenabled else projectdescriptioneditor.CustomAction("Enable developer documentation",
                                                                           self._enable_devdoc)]

    def _disable_devdoc(self, ais):
        for x in list(ais.entries()):
            if isinstance(x, anise.framework.projects.IntermediateStructure.Let):
                if x.key in ["documentation.has_developerdocumentation",
                             "documentation.readme_include_developerdocumentation"]:
                    ais.removeentry(x)

    def _enable_devdoc(self, ais):
        if not universe.ui.userfeedback.messageyesno(
                "Do you want to add developer documentation support to this project?"):
            return
        newlet = anise.framework.projects.IntermediateStructure.Let("documentation.has_developerdocumentation")
        ais.addentry(newlet)
        if universe.ui.userfeedback.messageyesno(
                "Do you also want to include the developer documentation in the user manual?"):
            newlet.setarg("also_include_in_readme", anise.framework.projects.IntermediateStructure.Value.frombool(True))
