# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for software tests as part of quality control.
"""

from typing import *
import datetime
import itertools
import os
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

releasing = anise.framework.features.loadfeature("releasing").featuremodule
uiwebhelpers = anise.framework.features.loadfeature("ui.web.helpers").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_DEFINITION)
    def _initproject():
        universe.testing.tests = Internals.Tests()

    @staticmethod
    @anise.framework.features.hookhandler(anise.framework.features.HOOK_BEFORE_EXECUTION)
    def _initproject2():
        if len(universe.testing.tests.list) > 0:
            universe.addtask(test, "testing.")

    class TestRun:
        """
        This instance keeps track of the test results. It is used by test runners for return test results.
        """

        #: Failed indicator.
        SUCCESS_FAILED = 0

        #: Success with warnings indicator.
        SUCCESS_SUCCESSFUL_WITH_WARNINGS = 1

        #: Success indicator.
        SUCCESS_SUCCESSFUL = 2

        def __init__(self):
            self.testpackages = {}
            self.testpackagelist = []
            self.currentpackagename = None
            self.currenttestname = None
            self.currentbegan = None

        def addresult(self, packagename: str, testname: str, output: str, success: int, duration_msec: float) -> None:
            """
            Adds a test result.

            :param packagename: The test package name.
            :param testname: The test name.
            :param output: The test run output.
            :param success: The test run success state.
            :param duration_msec: The execution duration in milliseconds.
            """
            if packagename not in self.testpackagelist:
                self.testpackages[packagename] = []
                self.testpackagelist.append(packagename)
            self.testpackages[packagename].append((testname, output, success, duration_msec))

        def testscount(self, with_success: Optional[int] = None) -> int:
            """
            Returns the number of tests done.

            :param with_success: Only count tests with this success state.
            """
            return len([x for x in itertools.chain.from_iterable(self.testpackages.values())
                        if ((with_success is None) or (with_success == x[2]))])

        def settest(self, packagename: str, testname: str) -> None:
            """
            Running setpackage, begintest and endtest together is a convenient alternative to addresult doing mostly the
            same.

            :param packagename: The test package name.
            :param testname: The test name.
            """
            self.currentpackagename = packagename
            self.currenttestname = testname

        def begintest(self) -> None:
            """
            Running setpackage, begintest and endtest together is a convenient alternative to addresult doing mostly the
            same.
            """
            self.currentbegan = datetime.datetime.now()

        def endtest(self, output: str, success: int) -> None:
            """
            Running setpackage, begintest and endtest together is a convenient alternative to addresult doing mostly the
            same.

            :param output: The test output.
            :param success: The test run success state.
            """
            duration_msec = (datetime.datetime.now() - self.currentbegan).total_seconds() * 1000
            self.addresult(self.currentpackagename, self.currenttestname, output, success, duration_msec)

    class TestReportWebModule(uiwebhelpers.AniseWebApplication):
        """
        Anise web module for generating the html test report.
        """

        scripts = """
            yaji.ready(function(){
                yaji.mainview.setBodyRight({view: 'HtmlView', hstretch: 1, vstretch: 1, 
                                            contentRoot: $(yaji.appconfig.getConfig('content'))[0]});
                yaji.mainview.setHead2({view: 'HtmlView', hstretch: 1, vstretch: 1, 
                                            contentRoot: $(yaji.appconfig.getConfig('header'))[0]});
            });
        """

        style = """
            #mainpart {
                position: absolute;
                top: 0;
                left: 0;
                bottom: 0;
                right: 0;
                overflow-y: auto;
                background-color: white;
                padding: 10pt;
            }
            .output h4 {
                font-size: 0.8em;
            }
            .scroller {
                background: rgba(0,0,0,0.05);
                overflow: auto;
                white-space: pre;
                font-family: monospace;
            }
            div.test {
                padding-left: 20pt;
                padding-bottom: 20pt;
            }
            .failed {
                color: red;
            }
            .successwithwarnings {
                color: olive;
            }
            .success {
                color: green;
            }
            .failed .scroller {
                color: #500;
                max-height: 400px;
            }
            .successwithwarnings .scroller {
                color: #440;
                max-height: 200px;
            }
            .success .scroller {
                color: #030;
                max-height: 100px;
            }
            .gauge {
                margin-bottom:5pt;
            }
            table.gauge {
                width: 300pt;
                height: 10pt;
                border-collapse: collapse;
                border: 1px solid gray;
            }
            table.gauge .left {
                background-color: #7f7;
            }
            table.gauge .right {
                background-color: #d33;
            }
            .package table.gauge .left {
                background-color: #afa;
            }
            .package table.gauge .right {
                background-color: #faa;
            }
        """

        htmlmainpart = """<div id="mainpart" class='{ffullsuccclass}'>{body}</div>"""

        headeradditions = "<div><div>{ffullsuccess} (in {fullduration_msec:.0f}ms) at {fdate}.</div>{fullgauge}</div>"

        _htmltemplate_testpackage = """
            <div class='{fpkgsuccclass} package'>
                <h2>test package '{testpackage}'</h2>
                <div>{fpkgsuccess} (in {packageduration_msec:.0f}ms).</div>
                {packagegauge}
                {testdetails}
            </div>
        """

        _htmltemplate_test = """
            <div class='{fsuccclass} test'>
                <h3>test '{testname}'</h3>
                <div>{fsuccess} (in {duration_msec:.0f}ms).</div>
                <div class='output'>
                    <h4>Output</h4>
                    <div class='scroller'>{output}</div>
                </div>
            </div>
        """

        @staticmethod
        def _successpresentation(_success, failcounter=None):
            if failcounter and failcounter[1] > 1:
                cmsg1 = " in {0} of {1} {2}".format(*failcounter)
                cmsg2 = " in {1} {2}".format(*failcounter)
            else:
                cmsg1 = cmsg2 = ""
            if _success == Internals.TestRun.SUCCESS_FAILED:
                return "Ran into <span class='failed'>problems</span>" + cmsg1, "failed", True
            elif _success == Internals.TestRun.SUCCESS_SUCCESSFUL_WITH_WARNINGS:
                return "Ran successfully" + cmsg2 + \
                       ", but <span class='successwithwarnings'>with warnings</span> in output", \
                       "successwithwarnings", True
            elif _success == Internals.TestRun.SUCCESS_SUCCESSFUL:
                return "Ran <span class='success'>successfully</span>" + cmsg2, "success", True
            else:
                return "Ran with unknown success due to backend error", "failed", False

        @staticmethod
        def _gauge(failed, total):
            if failed == 0 or total <= 1:
                return "<div class='gauge'></div>"
            else:
                return """
                    <table class='gauge'>
                        <tr>
                            <td class='left' style='width:{0}%;'></td>
                            <td class='right' style='width:{1}%;'></td>
                        </tr>
                    </table>
                """.format((total-failed)*100.0/total, failed*100.0/total)

        def __init__(self, testrun: 'Internals.TestRun', title: str):
            super().__init__(head1=title)
            self.testrun = testrun
            self.add_clientscript(Internals.TestReportWebModule.scripts)
            self.add_clientstyle(Internals.TestReportWebModule.style)

        def oninitialize(self):
            super().oninitialize()
            body = ""
            fullsuccess = Internals.TestRun.SUCCESS_SUCCESSFUL
            fullduration_msec = 0
            fullerrorcount = 0
            for testpackage in self.testrun.testpackagelist:
                testdetails = ""
                packagesuccess = Internals.TestRun.SUCCESS_SUCCESSFUL
                packageduration_msec = 0
                packageerrorcount = 0
                for testname, output, success, duration_msec in self.testrun.testpackages[testpackage]:
                    fsuccess, fsuccclass, understood = Internals.TestReportWebModule._successpresentation(success)
                    testdetails += Internals.TestReportWebModule._htmltemplate_test.format(**locals())
                    if understood:
                        packagesuccess = min(success, packagesuccess)
                        if success <= Internals.TestRun.SUCCESS_FAILED:
                            packageerrorcount += 1
                    else:
                        packagesuccess = Internals.TestRun.SUCCESS_FAILED
                        packageerrorcount += 1
                    packageduration_msec += duration_msec
                packagetestcount = len(self.testrun.testpackages[testpackage])
                fpkgsuccess, fpkgsuccclass, understood = Internals.TestReportWebModule._successpresentation(
                    packagesuccess, failcounter=(packageerrorcount, packagetestcount, "tests"))
                packagegauge = Internals.TestReportWebModule._gauge(packageerrorcount, packagetestcount)
                body += Internals.TestReportWebModule._htmltemplate_testpackage.format(**locals())
                fullsuccess = min(packagesuccess, fullsuccess)
                fullduration_msec += packageduration_msec
                if packagesuccess <= Internals.TestRun.SUCCESS_FAILED:
                    fullerrorcount += 1
            fullpackagecount = len(self.testrun.testpackagelist)
            fullgauge = Internals.TestReportWebModule._gauge(fullerrorcount, fullpackagecount)
            fdate = datetime.datetime.now().strftime("%c")
            ffullsuccess, ffullsuccclass, understood = Internals.TestReportWebModule._successpresentation(
                fullsuccess, failcounter=(fullerrorcount, fullpackagecount, "test packages"))
            self.appconfig.setconfig("content", Internals.TestReportWebModule.htmlmainpart.format(**locals()))
            self.appconfig.setconfig("header", Internals.TestReportWebModule.headeradditions.format(**locals()))

    @staticmethod
    def storereporthtml(testrun: 'Internals.TestRun', title: str, targetpath: str) -> str:
        """
        Stores the test report as html to filesystem.
        """
        Internals.TestReportWebModule(testrun, title).storeasstaticapplication(targetpath)
        os.makedirs(targetpath+"/static/icons/pde")
        shutil.copyfile(universe.specialpaths.anisedatadir+"/icons/pde/anise.png",
                        targetpath+"/static/icons/pde/anise.png")
        return "index.html"

    @staticmethod
    def storereporttext(testrun: 'Internals.TestRun', title: str, targetpath: str) -> str:
        """
        Stores the test report as plaintext to filesystem.
        """
        _template_testpackage = """
            test package '{testpackage}'
            {fpkgsuccess} (in {packageduration_msec:.0f}ms).
            {packagegauge}

{testdetails}
        """
        _template_test = """
        test '{testname}'
        {fsuccess} (in {duration_msec:.0f}ms).

    Output

{output}
        """
        _template_body = """
                {title}
                {ffullsuccess} (in {fullduration_msec:.0f}ms)
                at {fdate}.
                {fullgauge}

{body}
        """

        def _successpresentation(_success, failcounter=None):
            if failcounter and failcounter[1] > 1:
                cmsg1 = " in {0} of {1} {2}".format(*failcounter)
                cmsg2 = " in {1} {2}".format(*failcounter)
            else:
                cmsg1 = cmsg2 = ""
            if _success == Internals.TestRun.SUCCESS_FAILED:
                return "Ran into PROBLEMS" + cmsg1, "failed", True
            elif _success == Internals.TestRun.SUCCESS_SUCCESSFUL_WITH_WARNINGS:
                return "Ran successfully" + cmsg2 + ", but WITH WARNINGS in output", "successwithwarnings", True
            elif _success == Internals.TestRun.SUCCESS_SUCCESSFUL:
                return "Ran SUCCESSFULLY" + cmsg2, "success", True
            else:
                return "Ran with unknown success due to backend error", "failed", False

        def _gauge(failed, total):
            if failed == 0 or total <= 1:
                return ""
            else:
                failedparts = int(failed*30/total)
                goodparts = 30 - failedparts
                return "[{}{}]".format("|"*goodparts, " "*failedparts)
        body = ""
        fullsuccess = Internals.TestRun.SUCCESS_SUCCESSFUL
        fullduration_msec = 0
        fullerrorcount = 0
        for testpackage in testrun.testpackagelist:
            testdetails = ""
            packagesuccess = Internals.TestRun.SUCCESS_SUCCESSFUL
            packageduration_msec = 0
            packageerrorcount = 0
            for testname, output, success, duration_msec in testrun.testpackages[testpackage]:
                fsuccess, fsuccclass, understood = _successpresentation(success)
                testdetails += _template_test.format(**locals())
                if understood:
                    packagesuccess = min(success, packagesuccess)
                    if success <= Internals.TestRun.SUCCESS_FAILED:
                        packageerrorcount += 1
                else:
                    packagesuccess = Internals.TestRun.SUCCESS_FAILED
                    packageerrorcount += 1
                packageduration_msec += duration_msec
            packagetestcount = len(testrun.testpackages[testpackage])
            fpkgsuccess, fpkgsuccclass, understood = _successpresentation(packagesuccess,
                                                                          failcounter=(packageerrorcount,
                                                                                       packagetestcount, "tests"))
            packagegauge = _gauge(packageerrorcount, packagetestcount)
            body += _template_testpackage.format(**locals())
            fullsuccess = min(packagesuccess, fullsuccess)
            fullduration_msec += packageduration_msec
            if packagesuccess <= Internals.TestRun.SUCCESS_FAILED:
                fullerrorcount += 1
        fullpackagecount = len(testrun.testpackagelist)
        fullgauge = _gauge(fullerrorcount, fullpackagecount)
        fdate = datetime.datetime.now().strftime("%c")
        ffullsuccess, ffullsuccclass, understood = _successpresentation(fullsuccess,
                                                                        failcounter=(fullerrorcount, fullpackagecount,
                                                                                     "test packages"))
        fname = "testreport.txt"
        with open(targetpath+"/"+fname, "w") as reportfile:
            reportfile.write(_template_body.format(**locals()))
        return fname

    class Tests:
        """
        Storage of all tests.
        """

        def __init__(self):
            self.list = []

        def add(self, packagename: str, name: str, runner: Callable, testonrelease: bool = False, **kwargs):
            """
            Adds a test.

            :param packagename: The test package name.
            :param name: The test name.
            :param runner: The function to execute for running the test.
            :param testonrelease: If this test should automatically be made  a part of the releasing.
            :param kwargs: Additional arguments for the test runner function.
            """
            self.list.append((packagename, name, runner, kwargs))
            if testonrelease:
                self.test_on_release(packagename, name)

        def packagelist(self) -> List[str]:
            """
            Gets the list of all test package names.
            """
            return sorted(list(set([x[0] for x in self.list])))

        def listbypackage(self, package: str) -> List[Tuple[str, Callable, Dict[str, Any]]]:
            """
            Gets the list of all test names in a package.

            :param package: The test package name.
            """
            return [(n, r, kwargs) for pn, n, r, kwargs in self.list if pn == package]

        def test_on_release(self, package: str, name: str) -> None:
            """
            Adds a test to the releasing, so it is automatically executed each time :samp:`release` is executed.

            :param package: The test package name.
            :param name: The test name.
            """
            def _testonrelease():
                def _f():
                    r = universe.testing.test(runtestpackage=package, runtest=name)
                    if not r.testsuccess:
                        raise anise.framework.exceptions.RequirementsMissingError("a release-critical test failed")
                return _f
            universe.addhook(releasing.HOOK_RELEASE, _testonrelease(), provides=["testing", "PRE"])


def test(runtestpackage: Optional[str] = None,
         runtest: Optional[str] = None) -> Optional[anise.framework.files.Filestructure]:
    """
    Runs tests. It either obeys runtestpackage and runtest, checks the universe, or it asks the user which tests to run.

    :param runtestpackage: Name of the test package to run (:samp:`'*'`: all).
    :param runtest: Name of the test to run (:samp:`'*'`: all).
    """
    testrun = Internals.TestRun()
    testpackages = universe.testing.tests.packagelist()
    if not runtestpackage:
        runtestpackage = getattr(universe, "runtestpackage", None)
    if not runtestpackage:
        r = universe.ui.userfeedback.choice("Please select a test package.", ["All packages"] + testpackages)
        if r is None:
            return
        if r == 0:
            runtestpackage = "*"
        else:
            runtestpackage = testpackages[r-1]
    if runtestpackage != "*":
        tests = [x[0] for x in universe.testing.tests.listbypackage(runtestpackage)]
    else:
        tests = list(itertools.chain.from_iterable([[x[0] for x in universe.testing.tests.listbypackage(testpackage)]
                                                    for testpackage in universe.testing.tests.packagelist()]))
    if not runtest:
        runtest = getattr(universe, "runtest", None)
    if not runtest:
        r = universe.ui.userfeedback.choice("Please select a test.", ["All tests"] + tests)
        if r is None:
            return
        if r == 0:
            runtest = "*"
        else:
            runtest = tests[r-1]
    fname = universe.name
    fsubsetof = "subset of " if (runtest != "*") else ""
    if runtestpackage != "*":
        testpackages = [runtestpackage]
        title = f"Anise test report: {fname}: {fsubsetof}'{runtestpackage}'"
    else:
        title = f"Anise test report: {fname}: {fsubsetof}all"
    for testpackage in testpackages:
        for testname, runner, kwargs in universe.testing.tests.listbypackage(testpackage):
            if (runtest == "*") or (runtest == testname):
                testrun.settest(testpackage, testname)
                with universe.executionscope.newsubscope(f"Running test '{testpackage}:{testname}'"):
                    cntwbef = testrun.testscount(Internals.TestRun.SUCCESS_SUCCESSFUL_WITH_WARNINGS)
                    cntebef = testrun.testscount(Internals.TestRun.SUCCESS_FAILED)
                    runner(testrun, **kwargs)
                    cntw = testrun.testscount(Internals.TestRun.SUCCESS_SUCCESSFUL_WITH_WARNINGS) - cntwbef
                    cnte = testrun.testscount(Internals.TestRun.SUCCESS_FAILED) - cntebef
                    msgs = []
                    if cntw == 1:
                        msgs.append("1 warning")
                    elif cntw > 1:
                        msgs.append("{} warnings".format(cntw))
                    if cnte == 1:
                        msgs.append("1 failure")
                    elif cnte > 1:
                        msgs.append("{} failures".format(cnte))
                    if msgs:
                        anise.utils.logging.logwarning(", ".join(msgs), printcaller=False)
    ptestpackagename = runtestpackage if runtestpackage != "*" else "ALL"
    ptestname = runtest if runtest != "*" else "ALL"
    reportdir = anise.framework.files.TempDir(dirname=f"{universe.name}-testreport-{ptestpackagename}-{ptestname}")
    reportdir.testsuccess = len([x for x in itertools.chain.from_iterable([testrun.testpackages[x] for x in
                                                                           testrun.testpackages])
                                 if x[2] < Internals.TestRun.SUCCESS_SUCCESSFUL]) == 0
    if universe.ui.mode == universe.ui.Mode.WEB:
        reportfilename = Internals.storereporthtml(testrun, title, reportdir.path)
        interminal = False
    else:
        reportfilename = Internals.storereporttext(testrun, title, reportdir.path)
        interminal = True
    reportdir.setautoopen(f"/{reportfilename}", interminal=interminal)
    return reportdir
