# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
Features for interaction with the Git version control system.
"""

from typing import *
import datetime
import os
import re
import shutil

import anise.framework.exceptions
import anise.framework.features
import anise.framework.files
import anise.utils

from anise.framework.projects import universe

pdeversioncontrol = anise.framework.features.loadfeature("projectdescriptioneditor.versioncontrol").featuremodule
ui = anise.framework.features.loadfeature("ui").featuremodule
versioncontrol = anise.framework.features.loadfeature("versioncontrol").featuremodule


class Internals:

    @staticmethod
    @anise.framework.features.hookhandler(pdeversioncontrol.HOOK_GET_VERSIONCONTROLSYSTEM_IMPLEMENTATIONS)
    def _register_vcs():
        return GitVersionControlSystem, "versioncontrol.git", "Git"

    class Version:
        """
        Version number of an Anise Git project.
        """

        def __init__(self, build: int, tag: str):
            self.build = build
            self.tag = tag

        def isvalid(self) -> bool:
            return self.tag != "unknown"

        def __str__(self):
            return str(self.build) + (f"+{self.tag}" if self.tag else "")

    @staticmethod
    def callgit(*opts: str, raise_on_errors: Optional[str] = "errors in git", **kwargs) -> Tuple[int, Any]:
        gitopts = ["-c", "http.sslVerify=false"] if getattr(universe, "gitnohttpsslverify", False) else []
        return anise.utils.basic.call("git", *(gitopts + list(opts)), **kwargs, raise_on_errors=raise_on_errors)

    @staticmethod
    def getgitrevision(dirpath: str) -> 'Internals.Version':
        """
        Determines the revision number of the git working copy.

        :param dirpath: The working copy directory.
        """
        try:
            with anise.utils.basic.ChDir(dirpath):
                o, s = Internals.callgit("rev-list", "HEAD", "--count")
            build = s
            tag = None  # no 'alike' support in git
        except anise.framework.exceptions.ProcessExecutionFailedError:
            build = "0"
            tag = "unknown"
        return Internals.Version(int(build), tag)

    @staticmethod
    def getgitrevisionhashfromgitrevision(dirpath: str, revision: int) -> str:
        """
        Determines the git revision hash (this is the value that is also used by Git for addressing revisions)
        for a git revision number (which is not really a part of Git versioning).

        :param dirpath: The working copy directory.
        :param revision: The revision number (an Anise concept that does not really exist in Git).
        """
        with anise.utils.basic.ChDir(dirpath):
            o, s = Internals.callgit("rev-list", "HEAD")
            lst = s.split("\n")
            return lst[len(lst)-int(revision)]

    @staticmethod
    def listgitrepo(dirpath: str) -> List[str]:
        """
        Determines the file tree of a git repository.

        :param dirpath: The directory you want to list.
        """
        with anise.utils.basic.ChDir(dirpath):
            return [x for x in [a.strip() for a in Internals.callgit("ls-files")[1].split("\n")] if x]


# noinspection PyUnusedLocal
class GitVersionControlSystem(versioncontrol.Internals.VersionControlSystem):
    """
    Git version control system adapter.
    """

    def __init__(self):
        versioncontrol.Internals.VersionControlSystem.__init__(self)
        self.label = "Git"

    def getfullversion(self):
        _version = getattr(universe.base, "version", None)
        _revision = str(Internals.getgitrevision(universe.specialpaths.projectdir))
        return ((_version + ".") if _version is not None else "") + _revision

    def createrawpackage(self):
        with universe.executionscope.newsubscope("Fetching the sources from Git working copy"):
            tmp = anise.framework.files.TempDir("rawpackage")
            for e in Internals.listgitrepo(universe.specialpaths.projectdir):
                if e == ".gitignore":
                    continue
                full = f"{universe.specialpaths.projectdir}/{e}"
                tgtfull = f"{tmp.path}/{e}"
                if os.path.islink(full):
                    os.symlink(os.readlink(full), tgtfull)
                elif os.path.isfile(full):
                    if not os.path.exists(os.path.dirname(tgtfull)):
                        os.makedirs(os.path.dirname(tgtfull))
                    shutil.copy2(full, tgtfull)
                else:
                    raise anise.framework.exceptions.BadFormatError(f"unknown thing: {full}")
            return tmp

    def syncversioncontrolsystem(self, commitmessage="", forceskipped=True, forceunchanged=False):
        if commitmessage == "":
            commitmessage = "-"
        if forceskipped or (not getattr(universe, "_vcssynced", False)):
            anise.utils.logging.loginfo("Synchronizing the working copy with the git repository.", printcaller=False)
            with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
                Internals.callgit("add", "--update")
                Internals.callgit("pull")
                if forceunchanged or (Internals.callgit("diff", "--cached", "--exit-code",
                                                        raise_on_errors=None)[0] != 0):
                    Internals.callgit("commit", "--allow-empty-message", "--allow-empty", "-m", commitmessage)
                    Internals.callgit("push")
                    Internals.callgit("pull")
            universe._vcssynced = True

    def findvcs(self):
        return os.path.isdir(f"{universe.specialpaths.projectdir}/.git")

    def bindvcs(self, l):
        url = universe.ui.userfeedback.input("Please enter the git repository url.", f"https://foo/git/{universe.name}")
        if url is not None:
            with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
                if os.path.exists(f"{universe.specialpaths.projectdir}/.git"):
                    shutil.rmtree(f"{universe.specialpaths.projectdir}/.git")
                Internals.callgit("config", "credential.helper", "store")
                Internals.callgit("init")
                Internals.callgit("add", "--all")
                Internals.callgit("commit", "-m", "anise initial commit")
                Internals.callgit("remote", "add", "origin", url)
                Internals.callgit("push", "-u", "origin", "master")
                Internals.callgit("pull", "origin", "master")
                Internals.callgit("push", "-u", "origin", "master")
            return True

    def getheadrevision(self):
        return Internals.getgitrevision(universe.specialpaths.projectdir).build

    def fetchrawcommitmessage(self, revision):
        with anise.utils.basic.ChDir(universe.specialpaths.projectdir):
            gitrevhash = Internals.getgitrevisionhashfromgitrevision(universe.specialpaths.projectdir, revision)
            r, username = Internals.callgit("log", gitrevhash, "-n1", "--format=format:%cn")
            r, committime = Internals.callgit("log", gitrevhash, "-n1", "--format=format:%ct")
            r, commitsubject = Internals.callgit("log", gitrevhash, "-n1", "--format=format:%s")
            r, commitmsg = Internals.callgit("log", gitrevhash, "-n1", "--format=format:%b")
            return f"{revision}\n{username}\n{committime}\n{commitsubject}\n{commitmsg}"

    def getcommitmessagefromrawcommitmessage(self, s):
        rcommitmsg = [x.rstrip() for x in s.split("\n")]
        resrevision = rcommitmsg[0]
        resusername = rcommitmsg[1]
        rescommittime = datetime.datetime.fromtimestamp(int(rcommitmsg[2]))
        resmessage = "\n".join(rcommitmsg[3:])
        if resmessage == "-":
            resmessage = ""
        return versioncontrol.CommitMessage(resmessage, resrevision, rescommittime, resusername)

    def branchfromhere(self, branchname=None):
        if not branchname:
            branchname = universe.ui.userfeedback.input("Please enter the new branch name.", "")
        if branchname:
            Internals.callgit("checkout", "-b", branchname)
            _sbt = [x for x in Internals.callgit("show-branch", "-r")[1].split("\n") if x][-1]
            remotename = re.search(r"\[(.*)/.*\]", _sbt).group(1)
            Internals.callgit("push", "--set-upstream", remotename, branchname)
