# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
File structures, used for data exchange between methods or as result.
"""

from typing import *
import atexit
import os
import shutil
import sys
import tempfile

import anise.framework.exceptions
import anise.framework.globalvars
import anise.utils


class Filestructure:
    """
    This class represents a file or a directory structure with many subdirectories and files.
    There are methods for writing those items to disk and/or to modify the original structure
    in the filesystem.

    It is more powerful for file exchange than working just with file paths, because a file structure can also
    carry arbitrary metadata information and allows dynamic content creation (see e.g. :py:class:`TaskExecution`).

    Instances of this class (and its subclasses) are used by many features for file exchange.
    """

    def __init__(self, path: Optional[str], relative: bool = False):
        """
        :param path: Path to the file or directory.
        :param relative: If path is relative to the project directory.
        """
        if relative and path:
            from anise.framework.projects import universe
            path = f"{universe.specialpaths.projectdir}/{path}"
        self._path = os.path.abspath(path) if (path is not None) else None
        self._data = {'autoopen': None, 'autoopenterminal': None}
        self._initialized = False

    def datakeys(self) -> List[str]:
        """
        Returns a list of keys for all stored metadata properties.
        """
        return list(self._data.keys())

    def setdata(self, k: str, v: Optional[Any]) -> None:
        """
        Sets a metadata property.

        :param k: The key name.
        :param v: The value.
        """
        self.initialize()
        self._data[k] = v

    def getdata(self, k: str, deflt: Optional[Any]) -> Optional[Any]:
        """
        Gets a metadata property.

        :param k: The key name.
        :param deflt: The default value (returned if the key does not exist).
        """
        self.initialize()
        return self._data.get(k, deflt)

    def initialize(self) -> None:
        """
        Initializes this Filestructure. When this is called, the dynamic content (if any) is created and written
        to disc. It is automatically called for typical usage and is not required to be called manually.
        """
        if self._initialized:
            return
        self._initialize()
        self._initialized = True

    def _initialize(self) -> None:
        """
        The internal initialization routine, which fetches dynamic content (if any). Called from outside before
        actual accesses occur.
        |dooverride_optional|
        """
        pass

    @property
    def path(self) -> str:
        """
        The path to this file structure in the filesystem.
        """
        self.initialize()
        return self._path

    @property
    def data(self) -> Dict[str, Any]:
        """
        The file structure metadata.
        """
        self.initialize()
        return self._data

    def dl(self, to: Optional[str] = None, *, subpath: str = "",
           progresscallback: Optional[Callable[[int, int], None]] = None) -> str:
        """
        Copies the complete filestructure or a subdirectory to a new destination.

        :param to: If given, it specifies the destination path; if not given, a new directory in the current working
                   directory will be used.
        :param subpath: If given, only this subdirectory will be copied.
        :param progresscallback: Callback function for retrieving progress information.
        :return: The destination path.
        """
        self.initialize()
        # prepare destination
        if not to:
            to = f"{os.getcwd()}/{os.path.basename(self.path)}"
            if os.path.exists(to):  # we don't do that if we have an explicit destination
                anise.utils.logging.logwarning("removing old "+to, printcaller=False)
                if os.path.isdir(to):
                    shutil.rmtree(to)
                else:
                    os.unlink(to)
        # copy
        fullpath = os.path.abspath(f"{self.path}/{subpath}")
        if os.path.isfile(fullpath):
            dto = os.path.dirname(to)
            anise.utils.basic.mkdirs(dto)
            files = [""]
        elif os.path.isdir(fullpath):
            anise.utils.basic.mkdirs(to)
            files = os.listdir(fullpath)
        else:
            raise anise.framework.exceptions.UnexpectedSituationError("does not exist: " + fullpath)
        ffiles = []
        while len(files) > 0:
            fn = files.pop(0)
            ff = os.path.abspath(f"{fullpath}/{fn}")
            d = os.path.abspath(f"{to}/{fn}")
            if os.path.islink(ff):
                os.symlink(os.readlink(ff), d)
            elif os.path.isdir(ff):
                anise.utils.basic.mkdirs(d)
                files += [f"{fn}/{x}" for x in os.listdir(ff)]
            elif os.path.isfile(ff):
                ffiles.append((ff, d, os.path.getsize(ff)))
            else:
                raise anise.framework.exceptions.UnexpectedSituationError("unknown thing: " + ff)
        ffullsize = 0
        for ff, d, fsize in ffiles:
            ffullsize += fsize
        fdonesize = 0
        freportedperc = None
        for ff, d, fsize in ffiles:
            if os.path.isdir(d):
                d = os.path.join(d, os.path.basename(ff))
            with open(ff, 'rb') as fsrc:
                with open(d, 'wb') as fdst:
                    while True:
                        buf = fsrc.read(16*1024)
                        if not buf:
                            break
                        fdst.write(buf)
                        fdonesize += len(buf)
                        fmyperc = int(100. * fdonesize / ffullsize)
                        if freportedperc != fmyperc:
                            freportedperc = fmyperc
                            if progresscallback:
                                progresscallback(ffullsize, fdonesize)
            shutil.copystat(ff, d)
        return to

    def with_modified_rootname(self, newname: str) -> 'Filestructure':
        """
        Returns a (temporary) file structure clone with a new directory name.

        :param newname: The name that the new root directory should get.
        """
        self.initialize()
        newtmp = TempDir()
        nname = f"{newtmp.path}/{newname}"
        self.dl(nname)
        return Filestructure(nname)

    def mv(self, newname: str) -> 'Filestructure':
        """
        Moves the file structure to a new path.
        Returns a new file structure resulting in a movement from the old one, but also updates this one internally.

        :param newname: The new name (absolute or relative).
        """
        self.initialize()
        newpath = newname
        if sys.platform == "win32":
            newpath = newpath.replace("\\", "/")
        if "/" not in newpath:
            newpath = f"../{newpath}"
        if not newpath.startswith("/"):
            newpath = f"{self.path}/{newpath}"
        newpath = os.path.abspath(newpath)
        os.rename(self.path, newpath)
        rs = Filestructure(newpath)
        self._path = rs.path
        return rs

    def setautoopen(self, path: str, interminal: bool = False) -> None:
        """
        Places a hint that requests the Anise engine to automatically open a file within this structure when
        it is returned to the user.

        :param path: The relative path to the file that should automatically be opened.
        :param interminal: If this hint is for terminal (non-graphical) usage.
        """
        self.initialize()
        if interminal:
            self.setdata("autoopenterminal", path)
        else:
            self.setdata("autoopen", path)


class TempDir(Filestructure):
    """
    A special file structure that automatically creates a temporary directory in background.

    Example:

    .. code-block:: python

      def sometask():
          mytmp = files.TempDir()
          with open(mytmp.path + "/somefile", "w") as f:
              f.write("some content for our new file")
          return mytmp
    """

    def __init__(self, dirname: Optional[str] = None, temprootpath: Optional[str] = None,
                 namepattern: str = "anise.{pid}.{i}"):
        """
        :param dirname: An optional name that the root directory should get; if not given, it gets an random name.
        :param temprootpath: An optional alternate temp root directory path.
        :param namepattern: The directory name pattern.
        """
        i = 0
        path = None
        _pt = f"{temprootpath or tempfile.gettempdir()}/{namepattern}"
        pid = os.getpid()
        while path is None:
            _p = _pt.format(**locals())
            i += 1
            try:
                if not os.path.exists(_p):
                    os.mkdir(_p)
                    path = _p
                    atexit.register(self.__cleanup)
            except Exception:
                pass
        self.origpath = path
        if dirname:
            path += f"/{dirname}"
            os.mkdir(path)
        super().__init__(path)

    def __enter__(self):
        return self

    def mv(self, *a, **b):
        super().mv(*a, **b)
        self.origpath = None

    def __cleanup(self) -> None:
        """
        Deletes the original temp dir from the hard drive.
        """
        if anise.framework.globalvars.removetempdirs and self.origpath:
            try:
                shutil.rmtree(self.origpath)
            except Exception:
                pass

    def __exit__(self, etype, evalue, etraceback):
        self.__cleanup()


class TaskExecution(Filestructure):
    """
    Holds a function and some parameters for calling later on (like a delegate;
    or a more explicit, specialized and 'xml-compatible' version of a lambda) in order to
    get a file structure as function result. It acts as a proxy on top of this file structure.
    """

    def __init__(self, task: Callable, **kwargs):
        """
        :param task: The task implementation for execution.
        """
        super().__init__(None)
        self.__task = task
        self.__params = kwargs

    def _initialize(self):
        inner = self.__task(**self.__params)
        self._path = inner.path
        self._data = inner.data


class TextFileByContent(Filestructure):
    """
    Holds a function and some parameters for calling later on (like a delegate;
    or a more explicit, specialized and 'xml-compatible' version of a lambda) in order to
    get a file structure as function result. It acts as a proxy on top of this file structure.
    """

    def __init__(self, content: str):
        """
        :param content: The text content for the file.
        """
        fpath = f"{TempDir().path}/cnt"
        with open(fpath, "w") as f:
            f.write(content)
        super().__init__(fpath)


class AbstractRebuildDirectoriesFilestructure(TaskExecution):
    """
    Abstract base class for merging directory structures together. 
    
    See the subclasses.
    """

    def __init__(self):
        super().__init__(self.__getmerged)

    def getdests(self) -> List[Tuple[str, Filestructure]]:
        """
        Returns a list of transfer destinations.

        :return: A list of tuples, each beginning with the destination path (and other custom ones).
        """
        raise NotImplementedError()

    def getsource(self, desttuple: Tuple[str, Filestructure]) -> Filestructure:
        """
        Returns the transfer source for a destination tuple (as it was part of :py:meth:`getdests`' result).
        """
        raise NotImplementedError()

    def __getmerged(self) -> Filestructure:
        from anise.framework.projects import universe
        projectname = getattr(universe, "name", "aniseproject")
        result = TempDir()
        dests = self.getdests()
        for desttuple in sorted(dests, key=lambda aa: aa[0]):
            part = desttuple[0]
            fd = os.path.abspath(f"{result.path}/{part.format(name=projectname)}")
            anise.utils.basic.mkdirs(os.path.dirname(fd))
            srcfs = self.getsource(desttuple)
            if os.path.isdir(srcfs.path):
                srcfs.dl(fd)
            elif os.path.isfile(srcfs.path):
                shutil.copy2(srcfs.path, fd)
            else:
                raise anise.framework.exceptions.UnexpectedSituationError(f"unknown thing: {srcfs.path}")
        return result


class RemappedFilestructure(AbstractRebuildDirectoriesFilestructure):
    """
    Remaps a file structure's inner directory hierarchy into a new tree.
    """

    def __init__(self, source: Filestructure, frommap: Optional[Union[Dict[str, str], List[Tuple[str, str]]]] = None):
        """
        :param source: The source file structure.
        :param frommap: A dict of string/filestructure. Values are relative source paths with the dict key
                        specifying a target subpath.
        """
        super().__init__()
        self.__source = source
        self.__frommap = frommap or {}

    def getdests(self):
        res = self.__frommap
        if hasattr(res, "items"):
            res = res.items()
        return res

    def getsource(self, desttuple):
        return Filestructure(f"{self.__source.path}/{desttuple[1]}")


class MergedFilestructure(AbstractRebuildDirectoriesFilestructure):
    """
    Merges many file structures together in one tree.
    """

    def __init__(self, fromlist: Optional[List[Filestructure]] = None,
                 frommap: Optional[Union[Dict[str, Filestructure], List[Tuple[str, Filestructure]]]] = None):
        """
        :param fromlist: A list of file structures. Become root directory entries with their original file names.
        :param frommap: A dict of string/filestructure. Values become part of the result tree with the dict key
                        specifying a subpath.
        """
        super().__init__()
        self.__fromlist = fromlist or []
        self.__frommap = frommap or {}

    def getdests(self):
        res = self.__frommap
        if hasattr(res, "items"):
            res = res.items()
        res = list(res)
        for fl in self.__fromlist:
            res.append(("/", fl))
        return res

    def getsource(self, desttuple):
        return desttuple[1]


class ExcludeByName(TempDir):
    """
    Excludes some files/directories in a directory by name.
    """

    def __init__(self, source: Filestructure, names: List[str]):
        """
        :param source: The source file structure.
        :param names: Names to exclude.
        """
        self.__source = source
        self.__names = names
        super().__init__()

    def _initialize(self):
        super()._initialize()
        npath = f"{self._path}/x"
        self.__source.dl(npath)
        self._path = npath
        for fe in self.__names:
            ffe = f"{npath}/{fe}"
            if os.path.isdir(ffe):
                shutil.rmtree(ffe)
            elif os.path.isfile(ffe):
                os.unlink(ffe)
