# Copyright (C) 2014-2021, Josef Hahn
#
# This file is part of anise.
#
# anise is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# anise is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with anise.  If not, see <http://www.gnu.org/licenses/>.

"""
The Anise report facility keeps track of the steps of the execution process, the progresses and all output messages.
"""

from typing import *
import contextvars
import re
import sys
import threading

import anise.framework.exceptions
import anise.framework.features
import anise.framework.globalvars
import anise.utils


HOOK_EXECUTION_SCOPEPART_ADDED = anise.framework.features.Hook(
    doc="This hook is called when a new execution scope part appears.", wizard_functionparams=["newpart"])


class ExecutionScopePart:
    """
    Base class of :py:class:`ExecutionScope` (read there for more). It also includes logging output or output from
    external programs, which are displayed together with real execution scopes during task execution.
    """

    def __init__(self, value: str, hints: Optional[List[str]] = None):
        """
        :param value: The content value.
        :param hints: A list of hint symbols (typically used for log message severity).
        """
        self._address = []
        self._id = 0
        self._id_counter_ = 1
        self._scopepartsbyid = {}
        self._value = value
        self._hints = hints or []

    @property
    def id(self) -> int:
        """
        The ID of this part.
        """
        return self._id

    @property
    def value(self) -> str:
        """
        The content value of this execution scope.
        """
        return self._value

    @property
    def hints(self) -> List[str]:
        """
        The content hint.

        This is typically used for storing the severity of log messages, so the presentation can differ.
        """
        return self._hints

    @property
    def address(self) -> Tuple:
        """
        The address of this part.
        """
        return tuple(self._address)

    def getscopepartsbyidrange(self, minid: Optional[int] = None,
                               maxid: Optional[int] = None) -> List['ExecutionScopePart']:
        """
        Gets a list of execution scope parts by a range of IDs.
        """
        _lock.acquire()
        result = []
        if minid is None:
            minid = 0
        if maxid is None:
            maxid = max(rootscope._scopepartsbyid.keys())
        for i in range(minid, maxid+1):
            x = rootscope._scopepartsbyid.get(i)
            if x:
                result.append(x)
        _lock.release()
        return result


class ExecutionScope(ExecutionScopePart):
    """
    An execution scope is one piece of work to do in a complete task executions.

    It is the parent of another execution scope (just the root scope does not have a parent) and can have child scopes.
    This leads to a tree structure of scope nodes. Each node can have metadata infos like a title or a progress.
    Those are displayed during the execution of a task (together with other :py:class:`ExecutionScopePart` subclasses'
    instances).
    """

    class Progress:
        """
        Progress info for an execution scope.
        """

        def __init__(self):
            pass

        def tojson(self) -> Dict[str, Optional[Any]]:
            """
            Returns a simple object structure for json transfers.
            """
            return self.__dict__

    class ProgressDeterminate(Progress):
        """
        Progress info for a determinate progress.
        """

        def __init__(self, value: float):
            """
            :param value: The progress between :samp:`0.0` and :samp:`1.0`.
            """
            super().__init__()
            self.value = min(max(value, 0), 1)
            self.kind = "determinate"
            self.stopped = False

    class ProgressIndeterminate(Progress):
        """
        Progress info for an indeterminate progress.
        """

        def __init__(self):
            super().__init__()
            self.kind = "indeterminate"
            self.stopped = False

    class ProgressFinished(Progress):
        """
        Progress info for a finished execution.
        """

        def __init__(self):
            super().__init__()
            self.kind = "finished"
            self.stopped = True

    class ProgressDoneWithUnknownSuccess(Progress):
        """
        Progress info for an execution that is terminated but without explicit success information.
        """

        def __init__(self):
            super().__init__()
            self.kind = "donewithunknownsuccess"
            self.stopped = True

    class ProgressFailed(Progress):
        """
        Progress info for a failed execution.
        """

        def __init__(self):
            super().__init__()
            self.kind = "failed"
            self.stopped = True

    def __init__(self, title: str):
        """
        :param title: The scope title text.
        """
        super().__init__("")
        self._subparts = []
        self._progress = ExecutionScope.ProgressIndeterminate()
        self._title = title

    def appendchild(self, child: ExecutionScopePart) -> None:
        """
        Appends a child scope to this scope.
        """
        from anise.framework.projects import currentuniverse
        _lock.acquire()
        child._parentscope = self
        child._address = list(self._address+[len(self._subparts)])
        child._id = rootscope._id_counter_
        self._subparts.append(child)
        rootscope._scopepartsbyid[child.id] = child
        rootscope._id_counter_ += 1
        if currentuniverse():
            for h in currentuniverse().gethooks(HOOK_EXECUTION_SCOPEPART_ADDED):
                h(child)
        _lock.release()

    @property
    def progress(self) -> Progress:
        """
        The current progress state of this scope.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value

    @property
    def title(self) -> str:
        """
        The title string of this scope.
        """
        return self._title

    @property
    def rawsubparts(self) -> List['ExecutionScopePart']:
        """
        List of all direct subparts of this scope.
        """
        return list(self._subparts)

    def newsubscope(self, title: str) -> 'ExecutionScopePart':
        """
        Adds a new sub scope to the current top scope.

        :param title: The scope title text.
        """
        result = ExecutionScope(title=title)
        topscope().appendchild(result)
        return result

    def __enter__(self):
        if self._title:
            anise.utils.logging.loginfo(f"Begin: {self._title}", printcaller=False, terminalonly=True)
        _scopestack.set(_scopestack.get() + (self,))
        _globscopestack.append(self)
        return self

    def __exit__(self, etype, evalue, etraceback):
        _scopestack.set(_scopestack.get()[:-1])
        _globscopestack.pop()

        def _finishscope(sc):
            if etype:
                sc.progress = ExecutionScope.ProgressFailed()
            elif not isinstance(sc.progress, ExecutionScope.ProgressFailed):
                for csc in sc.rawsubparts:
                    if isinstance(csc, ExecutionScope):
                        _finishscope(csc)
                sc.progress = ExecutionScope.ProgressFinished()
        _finishscope(self)
        if self.title:
            if etype:
                anise.utils.logging.logerror(f"Failed: {self.title}", printcaller=False, terminalonly=True)
            else:
                anise.utils.logging.loginfo(f"Finished: {self.title}", printcaller=False, terminalonly=True)


_scopestack = contextvars.ContextVar('_scopestack', default=tuple())
_globscopestack = []


class RootExecutionScope(ExecutionScope):
    """
    The root execution scope.
    """

    def __init__(self):
        super().__init__(title="")


rootscope = RootExecutionScope()
rootscope.__enter__()

_lock = threading.Lock()


class AniseOutputStream:
    """
    Proxy for terminal output/error channel, so we can record output messages.
    """

    def __init__(self, originalstream: 'TextIO'):
        self.result = ""
        self.originalstream = originalstream

    def write(self, b):
        b = re.sub("\033\\[[0-9;]+m", "", b)
        topscope().appendchild(ExecutionScopePart(b))
        if self.originalstream:
            self.originalstream.write(b)

    def __getattr__(self, item):
        return getattr(self.originalstream, item)

    def __dir__(self):
        result = set(self.__dict__.keys())
        for k in dir(self.originalstream):
            result.add(k)
        return list(result)


def topscope(i: int = 0) -> ExecutionScope:
    """
    Returns a scope from the top of the stack (without removing it).

    :param i: Which scope to return (counting from top).
    """
    return (_scopestack.get() or _globscopestack)[-i-1]


def install_proxies() -> None:
    """
    Installs stdout/stderr proxies, so we can record all output made.
    """
    sys.stdout = AniseOutputStream(sys.stdout)
    sys.stderr = AniseOutputStream(sys.stderr)


def _logmessage(text: str, severity: int, terminalonly: bool, tofile: 'TextIO',
                textescaper: Callable[[str], str]) -> None:
    """
    A message logger implementation that also records the logging output to the report.

    :param text: The text to log.
    :param severity: The severity. One of :py:class:`anise.utils.logging.Severity`.
    :param terminalonly: If to log only to terminal/file.
    :param tofile: Log to this file instead of terminal.
    :param textescaper: A function that escapes text in a custom way before logging.
    """
    if (severity != anise.utils.logging.Severity.Debug) or anise.framework.globalvars.logdebug:
        rtext = ((len(_scopestack.get()) - 1) * ". ") + text
        print(textescaper(rtext), file=tofile)
    if not terminalonly:
        if severity == anise.utils.logging.Severity.Debug:
            hints = ["logdebug"]
        elif severity == anise.utils.logging.Severity.Info:
            hints = ["loginfo"]
        elif severity == anise.utils.logging.Severity.Warning:
            hints = ["logwarning"]
        elif severity == anise.utils.logging.Severity.Error:
            hints = ["logerror"]
        else:
            raise anise.framework.exceptions.BadInputError(f"unknown severity: {repr(severity)}")
        topscope().appendchild(ExecutionScopePart(text, hints=hints))


anise.utils.logging.setlowlevellogger(_logmessage)
